import logging
from collections import defaultdict
from typing import Any, List, Tuple

import pyarrow as pa
import ray
from ray import cloudpickle
from ray.types import ObjectRef

from deltacat import logs
from deltacat.compute.compactor import PrimaryKeyIndexVersionLocator, PyArrowWriteResult
from deltacat.compute.compactor.utils import primary_key_index as pki

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


@ray.remote(num_cpus=1, num_returns=2)
def rewrite_index(
    object_ids: List[Any],
    s3_bucket: str,
    new_primary_key_index_version_locator: PrimaryKeyIndexVersionLocator,
    max_records_per_index_file: int,
) -> Tuple[PyArrowWriteResult, List[ObjectRef]]:

    logger.info(f"Starting rewrite primary key index task...")
    object_refs = [cloudpickle.loads(obj_id_pkl) for obj_id_pkl in object_ids]
    logger.info(f"Getting table groups object refs...")
    table_groups_list = ray.get(object_refs)
    logger.info(f"Got {len(table_groups_list)} table groups object refs...")
    hb_index_to_tables = defaultdict(list)
    for table_groups in table_groups_list:
        for hb_index, table in enumerate(table_groups):
            if table is not None:
                hb_index_to_tables[hb_index].append(table)
    logger.info(f"Running {len(hb_index_to_tables)} rewrite index rounds...")
    pki_stats = []
    for hb_index, tables in hb_index_to_tables.items():
        table = pa.concat_tables(tables)
        hb_pki_stats = pki.write_primary_key_index_files(
            table,
            new_primary_key_index_version_locator,
            s3_bucket,
            hb_index,
            max_records_per_index_file,
        )
        pki_stats.append(hb_pki_stats)
    logger.info(f"Finished rewrite primary key index task...")
    return PyArrowWriteResult.union(pki_stats), object_refs



import logging
from typing import Callable, Dict, List, Optional

from fsspec import AbstractFileSystem
from pyarrow import csv as pacsv
from ray.data import Dataset
from ray.data.datasource import BlockWritePathProvider

from deltacat import logs
from deltacat.types.media import ContentEncoding, ContentType

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def write_parquet(
    dataset: Dataset,
    base_path: str,
    *,
    filesystem: AbstractFileSystem,
    block_path_provider: BlockWritePathProvider,
    **kwargs,
) -> None:

    dataset.write_parquet(
        base_path,
        filesystem=filesystem,
        try_create_dir=False,
        block_path_provider=block_path_provider,
        **kwargs,
    )


def write_csv(
    dataset: Dataset,
    base_path: str,
    *,
    filesystem: AbstractFileSystem,
    block_path_provider: BlockWritePathProvider,
    **kwargs,
) -> None:

    # column names are kept in table metadata, so omit header
    arrow_csv_args_fn = lambda: {
        "write_options": pacsv.WriteOptions(include_header=False)
    }
    pa_open_stream_args = {"compression": ContentEncoding.GZIP.value}
    dataset.write_csv(
        base_path,
        arrow_open_stream_args=pa_open_stream_args,
        filesystem=filesystem,
        try_create_dir=False,
        block_path_provider=block_path_provider,
        arrow_csv_args_fn=arrow_csv_args_fn,
        **kwargs,
    )


CONTENT_TYPE_TO_DATASET_WRITE_FUNC: Dict[str, Callable] = {
    ContentType.CSV.value: write_csv,
    ContentType.PARQUET.value: write_parquet,
}


def slice_dataset(dataset: Dataset, max_len: Optional[int]) -> List[Dataset]:
    """
    Returns equally-sized dataset slices of up to `max_len` records each.
    """
    # TODO (pdames): Since the purpose of this function is to ideally produce
    #  a dataset that won't write less than `max_len` records per file when the
    #  total dataset size is greater than `max_len`, we should also ensure that
    #  each dataset returned only contains 1 block.
    if max_len is None:
        return [dataset]
    dataset_len = dataset.count()
    num_splits = int(dataset_len / max_len)
    if dataset_len % max_len == 0:
        num_splits -= 1
    if num_splits <= 0:
        return [dataset]
    split_indices = [max_len + max_len * i for i in range(num_splits)]
    return dataset.split_at_indices(split_indices)


def dataset_size(dataset: Dataset) -> int:
    return dataset.size_bytes()


def dataset_to_file(
    table: Dataset,
    base_path: str,
    file_system: AbstractFileSystem,
    block_path_provider: BlockWritePathProvider,
    content_type: str = ContentType.PARQUET.value,
    **kwargs,
) -> None:
    """
    Writes the given Distributed Dataset to one or more files.
    """
    writer = CONTENT_TYPE_TO_DATASET_WRITE_FUNC.get(content_type)
    if not writer:
        raise NotImplementedError(
            f"Distributed Dataset writer for content type '{content_type}' not"
            f" implemented. Known content types: "
            f"{CONTENT_TYPE_TO_DATASET_WRITE_FUNC.keys}"
        )
    writer(
        table,
        base_path,
        filesystem=file_system,
        block_path_provider=block_path_provider,
        **kwargs,
    )



import logging
import time
from collections import defaultdict
from itertools import chain, repeat
from typing import List, Optional, Tuple

import pyarrow as pa
import ray
from ray import cloudpickle

from deltacat import logs
from deltacat.compute.compactor import (
    MaterializeResult,
    PyArrowWriteResult,
    RoundCompletionInfo,
)
from deltacat.compute.compactor.steps.dedupe import (
    DedupeTaskIndexWithObjectId,
    DeltaFileLocatorToRecords,
)
from deltacat.storage import Delta, DeltaLocator, Partition, PartitionLocator
from deltacat.storage import interface as unimplemented_deltacat_storage
from deltacat.types.media import DELIMITED_TEXT_CONTENT_TYPES, ContentType
from deltacat.types.tables import TABLE_CLASS_TO_SIZE_FUNC
from deltacat.utils.performance import timed_invocation
from deltacat.utils.pyarrow import (
    ReadKwargsProviderPyArrowCsvPureUtf8,
    ReadKwargsProviderPyArrowSchemaOverride,
    RecordBatchTables,
)

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


@ray.remote
def materialize(
    source_partition_locator: PartitionLocator,
    round_completion_info: Optional[RoundCompletionInfo],
    partition: Partition,
    mat_bucket_index: int,
    dedupe_task_idx_and_obj_id_tuples: List[DedupeTaskIndexWithObjectId],
    max_records_per_output_file: int,
    compacted_file_content_type: ContentType,
    schema: Optional[pa.Schema] = None,
    deltacat_storage=unimplemented_deltacat_storage,
) -> MaterializeResult:
    # TODO (rkenmi): Add docstrings for the steps in the compaction workflow
    #  https://github.com/ray-project/deltacat/issues/79
    def _materialize(compacted_tables: List[pa.Table]) -> MaterializeResult:
        compacted_table = pa.concat_tables(compacted_tables)

        if compacted_file_content_type in DELIMITED_TEXT_CONTENT_TYPES:
            # TODO (rkenmi): Investigate if we still need to convert this table to pandas DataFrame
            # TODO (pdames): compare performance to pandas-native materialize path
            df = compacted_table.to_pandas(
                split_blocks=True, self_destruct=True, zero_copy_only=True
            )
            compacted_table = df
        delta, stage_delta_time = timed_invocation(
            deltacat_storage.stage_delta,
            compacted_table,
            partition,
            max_records_per_entry=max_records_per_output_file,
            content_type=compacted_file_content_type,
        )
        compacted_table_size = TABLE_CLASS_TO_SIZE_FUNC[type(compacted_table)](
            compacted_table
        )
        logger.debug(
            f"Time taken for materialize task"
            f" to upload {len(compacted_table)} records"
            f" of size {compacted_table_size} is: {stage_delta_time}s"
        )
        manifest = delta.manifest
        manifest_records = manifest.meta.record_count
        assert (
            manifest_records == len(compacted_table),
            f"Unexpected Error: Materialized delta manifest record count "
            f"({manifest_records}) does not equal compacted table record count "
            f"({len(compacted_table)})",
        )
        materialize_result = MaterializeResult.of(
            delta,
            mat_bucket_index,
            # TODO (pdames): Generalize WriteResult to contain in-memory-table-type
            #  and in-memory-table-bytes instead of tight coupling to paBytes
            PyArrowWriteResult.of(
                len(manifest.entries),
                TABLE_CLASS_TO_SIZE_FUNC[type(compacted_table)](compacted_table),
                manifest.meta.content_length,
                len(compacted_table),
            ),
        )
        logger.info(f"Materialize result: {materialize_result}")
        return materialize_result

    logger.info(
        f"Starting materialize task with"
        f" materialize bucket index: {mat_bucket_index}..."
    )
    start = time.time()
    dedupe_task_idx_and_obj_ref_tuples = [
        (
            t1,
            cloudpickle.loads(t2),
        )
        for t1, t2 in dedupe_task_idx_and_obj_id_tuples
    ]
    logger.info(f"Resolved materialize task obj refs...")
    dedupe_task_indices, obj_refs = zip(*dedupe_task_idx_and_obj_ref_tuples)
    # this depends on `ray.get` result order matching input order, as per the
    # contract established in: https://github.com/ray-project/ray/pull/16763
    src_file_records_list = ray.get(list(obj_refs))
    all_src_file_records = defaultdict(list)
    for i, src_file_records in enumerate(src_file_records_list):
        dedupe_task_idx = dedupe_task_indices[i]
        for src_dfl, record_numbers in src_file_records.items():
            all_src_file_records[src_dfl].append(
                (record_numbers, repeat(dedupe_task_idx, len(record_numbers)))
            )
    manifest_cache = {}
    materialized_results: List[MaterializeResult] = []
    record_batch_tables = RecordBatchTables(max_records_per_output_file)
    for src_dfl in sorted(all_src_file_records.keys()):
        record_numbers_dd_task_idx_tpl_list: List[
            Tuple[DeltaFileLocatorToRecords, repeat]
        ] = all_src_file_records[src_dfl]
        record_numbers_tpl, dedupe_task_idx_iter_tpl = zip(
            *record_numbers_dd_task_idx_tpl_list
        )
        is_src_partition_file_np = src_dfl.is_source_delta
        src_stream_position_np = src_dfl.stream_position
        src_file_idx_np = src_dfl.file_index
        src_file_partition_locator = (
            source_partition_locator
            if is_src_partition_file_np
            else round_completion_info.compacted_delta_locator.partition_locator
        )
        delta_locator = DeltaLocator.of(
            src_file_partition_locator,
            src_stream_position_np.item(),
        )
        dl_digest = delta_locator.digest()

        manifest = manifest_cache.setdefault(
            dl_digest,
            deltacat_storage.get_delta_manifest(delta_locator),
        )

        read_kwargs_provider = None
        # for delimited text output, disable type inference to prevent
        # unintentional type-casting side-effects and improve performance
        if compacted_file_content_type in DELIMITED_TEXT_CONTENT_TYPES:
            read_kwargs_provider = ReadKwargsProviderPyArrowCsvPureUtf8()
        # enforce a consistent schema if provided, when reading files into PyArrow tables
        elif schema is not None:
            read_kwargs_provider = ReadKwargsProviderPyArrowSchemaOverride(
                schema=schema
            )
        pa_table, download_delta_manifest_entry_time = timed_invocation(
            deltacat_storage.download_delta_manifest_entry,
            Delta.of(delta_locator, None, None, None, manifest),
            src_file_idx_np.item(),
            file_reader_kwargs_provider=read_kwargs_provider,
        )
        logger.debug(
            f"Time taken for materialize task"
            f" to download delta locator {delta_locator} with entry ID {src_file_idx_np.item()}"
            f" is: {download_delta_manifest_entry_time}s"
        )
        mask_pylist = list(repeat(False, len(pa_table)))
        record_numbers = chain.from_iterable(record_numbers_tpl)
        # TODO(raghumdani): reference the same file URIs while writing the files
        # instead of copying the data over and creating new files.
        for record_number in record_numbers:
            mask_pylist[record_number] = True
        mask = pa.array(mask_pylist)
        pa_table = pa_table.filter(mask)
        record_batch_tables.append(pa_table)
        if record_batch_tables.has_batches():
            batched_tables = record_batch_tables.evict()
            materialized_results.append(_materialize(batched_tables))

    if record_batch_tables.has_remaining():
        materialized_results.append(_materialize(record_batch_tables.remaining))

    merged_delta = Delta.merge_deltas([mr.delta for mr in materialized_results])
    assert (
        materialized_results and len(materialized_results) > 0
    ), f"Expected at least one materialized result in materialize step."

    write_results = [mr.pyarrow_write_result for mr in materialized_results]
    logger.debug(
        f"{len(write_results)} files written"
        f" with records: {[wr.records for wr in write_results]}"
    )
    # Merge all new deltas into one for this materialize bucket index
    merged_materialize_result = MaterializeResult.of(
        merged_delta,
        materialized_results[0].task_index,
        PyArrowWriteResult.union(
            [mr.pyarrow_write_result for mr in materialized_results]
        ),
    )
    logger.info(f"Finished materialize task...")
    end = time.time()
    logger.info(f"Materialize task ended in {end - start}s")
    return merged_materialize_result



from deltacat.compute.compactor.model.delta_annotated import DeltaAnnotated
from deltacat.compute.compactor.model.delta_file_envelope import DeltaFileEnvelope
from deltacat.compute.compactor.model.delta_file_locator import DeltaFileLocator
from deltacat.compute.compactor.model.materialize_result import MaterializeResult
from deltacat.compute.compactor.model.primary_key_index import (
    PrimaryKeyIndexLocator,
    PrimaryKeyIndexMeta,
    PrimaryKeyIndexVersionLocator,
    PrimaryKeyIndexVersionMeta,
)
from deltacat.compute.compactor.model.pyarrow_write_result import PyArrowWriteResult
from deltacat.compute.compactor.model.round_completion_info import RoundCompletionInfo
from deltacat.compute.compactor.model.sort_key import SortKey, SortOrder

__all__ = [
    "DeltaAnnotated",
    "DeltaFileEnvelope",
    "DeltaFileLocator",
    "MaterializeResult",
    "PrimaryKeyIndexLocator",
    "PrimaryKeyIndexMeta",
    "PrimaryKeyIndexVersionLocator",
    "PrimaryKeyIndexVersionMeta",
    "PyArrowWriteResult",
    "RoundCompletionInfo",
    "SortKey",
    "SortOrder",
]



from itertools import repeat
from typing import Union

import numpy as np
import pyarrow as pa

from deltacat.compute.compactor import DeltaFileEnvelope
from deltacat.storage import DeltaType

_SYS_COL_UUID = "4000f124-dfbd-48c6-885b-7b22621a6d41"


def _get_sys_col_name(suffix):
    return f"{_SYS_COL_UUID}_{suffix}"


_PK_HASH_DIGEST_BYTE_WIDTH = 20
_PK_HASH_COLUMN_NAME = _get_sys_col_name("hash")
_PK_HASH_COLUMN_TYPE = pa.binary(_PK_HASH_DIGEST_BYTE_WIDTH)
_PK_HASH_COLUMN_FIELD = pa.field(
    _PK_HASH_COLUMN_NAME,
    _PK_HASH_COLUMN_TYPE,
)

_DEDUPE_TASK_IDX_COLUMN_NAME = _get_sys_col_name("dedupe_task_idx")
_DEDUPE_TASK_IDX_COLUMN_TYPE = pa.int32()
_DEDUPE_TASK_IDX_COLUMN_FIELD = pa.field(
    _DEDUPE_TASK_IDX_COLUMN_NAME,
    _DEDUPE_TASK_IDX_COLUMN_TYPE,
)

_PARTITION_STREAM_POSITION_COLUMN_NAME = _get_sys_col_name("stream_position")
_PARTITION_STREAM_POSITION_COLUMN_TYPE = pa.int64()
_PARTITION_STREAM_POSITION_COLUMN_FIELD = pa.field(
    _PARTITION_STREAM_POSITION_COLUMN_NAME,
    _PARTITION_STREAM_POSITION_COLUMN_TYPE,
)

_ORDERED_FILE_IDX_COLUMN_NAME = _get_sys_col_name("file_index")
_ORDERED_FILE_IDX_COLUMN_TYPE = pa.int32()
_ORDERED_FILE_IDX_COLUMN_FIELD = pa.field(
    _ORDERED_FILE_IDX_COLUMN_NAME,
    _ORDERED_FILE_IDX_COLUMN_TYPE,
)

_ORDERED_RECORD_IDX_COLUMN_NAME = _get_sys_col_name("record_index")
_ORDERED_RECORD_IDX_COLUMN_TYPE = pa.int64()
_ORDERED_RECORD_IDX_COLUMN_FIELD = pa.field(
    _ORDERED_RECORD_IDX_COLUMN_NAME,
    _ORDERED_RECORD_IDX_COLUMN_TYPE,
)

_DELTA_TYPE_COLUMN_NAME = _get_sys_col_name("delta_type")
_DELTA_TYPE_COLUMN_TYPE = pa.bool_()
_DELTA_TYPE_COLUMN_FIELD = pa.field(
    _DELTA_TYPE_COLUMN_NAME,
    _DELTA_TYPE_COLUMN_TYPE,
)

_IS_SOURCE_COLUMN_NAME = _get_sys_col_name("is_source")
_IS_SOURCE_COLUMN_TYPE = pa.bool_()
_IS_SOURCE_COLUMN_FIELD = pa.field(
    _IS_SOURCE_COLUMN_NAME,
    _IS_SOURCE_COLUMN_TYPE,
)


def get_pk_hash_column_array(obj) -> Union[pa.Array, pa.ChunkedArray]:
    return pa.array(obj, _PK_HASH_COLUMN_TYPE)


def pk_hash_column_np(table: pa.Table) -> np.ndarray:
    return table[_PK_HASH_COLUMN_NAME].to_numpy()


def pk_hash_column(table: pa.Table) -> pa.ChunkedArray:
    return table[_PK_HASH_COLUMN_NAME]


def delta_type_column_np(table: pa.Table) -> np.ndarray:
    return table[_DELTA_TYPE_COLUMN_NAME].to_numpy()


def delta_type_column(table: pa.Table) -> pa.ChunkedArray:
    return table[_DELTA_TYPE_COLUMN_NAME]


def get_dedupe_task_idx_column_array(obj) -> Union[pa.Array, pa.ChunkedArray]:
    return pa.array(
        obj,
        _DEDUPE_TASK_IDX_COLUMN_TYPE,
    )


def get_stream_position_column_array(obj) -> Union[pa.Array, pa.ChunkedArray]:
    return pa.array(
        obj,
        _PARTITION_STREAM_POSITION_COLUMN_TYPE,
    )


def stream_position_column_np(table: pa.Table) -> np.ndarray:
    return table[_PARTITION_STREAM_POSITION_COLUMN_NAME].to_numpy()


def get_file_index_column_array(obj) -> Union[pa.Array, pa.ChunkedArray]:
    return pa.array(
        obj,
        _ORDERED_FILE_IDX_COLUMN_TYPE,
    )


def file_index_column_np(table: pa.Table) -> np.ndarray:
    return table[_ORDERED_FILE_IDX_COLUMN_NAME].to_numpy()


def get_record_index_column_array(obj) -> Union[pa.Array, pa.ChunkedArray]:
    return pa.array(
        obj,
        _ORDERED_RECORD_IDX_COLUMN_TYPE,
    )


def record_index_column_np(table: pa.Table) -> np.ndarray:
    return table[_ORDERED_RECORD_IDX_COLUMN_NAME].to_numpy()


def is_source_column_np(table: pa.Table) -> np.ndarray:
    return table[_IS_SOURCE_COLUMN_NAME].to_numpy()


def get_delta_type_column_array(obj) -> Union[pa.Array, pa.ChunkedArray]:
    return pa.array(
        obj,
        _DELTA_TYPE_COLUMN_TYPE,
    )


def get_is_source_column_array(obj) -> Union[pa.Array, pa.ChunkedArray]:
    return pa.array(
        obj,
        _IS_SOURCE_COLUMN_TYPE,
    )


def project_delta_file_metadata_on_table(
    delta_file_envelope: DeltaFileEnvelope,
) -> pa.Table:

    table = delta_file_envelope.table

    # append ordered file number column
    ordered_file_number = delta_file_envelope.file_index
    ordered_file_number_iterator = repeat(
        int(ordered_file_number),
        len(table),
    )
    table = append_file_idx_column(table, ordered_file_number_iterator)

    # append event timestamp column
    stream_position = delta_file_envelope.stream_position
    stream_position_iterator = repeat(
        int(stream_position),
        len(table),
    )
    table = append_stream_position_column(table, stream_position_iterator)

    # append delta type column
    delta_type = delta_file_envelope.delta_type
    delta_type_iterator = repeat(
        delta_type_to_field(delta_type),
        len(table),
    )
    table = append_delta_type_col(table, delta_type_iterator)

    # append is source column
    is_source_iterator = repeat(
        True,
        len(table),
    )
    table = append_is_source_col(table, is_source_iterator)
    return table


def append_stream_position_column(table: pa.Table, stream_positions):

    table = table.append_column(
        _PARTITION_STREAM_POSITION_COLUMN_FIELD,
        get_stream_position_column_array(stream_positions),
    )
    return table


def append_file_idx_column(table: pa.Table, ordered_file_indices):

    table = table.append_column(
        _ORDERED_FILE_IDX_COLUMN_FIELD,
        get_file_index_column_array(ordered_file_indices),
    )
    return table


def append_pk_hash_column(table: pa.Table, pk_hashes) -> pa.Table:

    table = table.append_column(
        _PK_HASH_COLUMN_FIELD, get_pk_hash_column_array(pk_hashes)
    )
    return table


def append_record_idx_col(table: pa.Table, ordered_record_indices) -> pa.Table:

    table = table.append_column(
        _ORDERED_RECORD_IDX_COLUMN_FIELD,
        get_record_index_column_array(ordered_record_indices),
    )
    return table


def append_dedupe_task_idx_col(table: pa.Table, dedupe_task_indices) -> pa.Table:

    table = table.append_column(
        _DEDUPE_TASK_IDX_COLUMN_FIELD,
        get_dedupe_task_idx_column_array(dedupe_task_indices),
    )
    return table


def delta_type_to_field(delta_type: DeltaType) -> bool:
    return True if delta_type is DeltaType.UPSERT else False


def delta_type_from_field(delta_type_field: bool) -> DeltaType:
    return DeltaType.UPSERT if delta_type_field else DeltaType.DELETE


def append_delta_type_col(table: pa.Table, delta_types) -> pa.Table:

    table = table.append_column(
        _DELTA_TYPE_COLUMN_FIELD,
        get_delta_type_column_array(delta_types),
    )
    return table


def append_is_source_col(table: pa.Table, booleans) -> pa.Table:

    table = table.append_column(
        _IS_SOURCE_COLUMN_FIELD,
        get_is_source_column_array(booleans),
    )
    return table


def get_minimal_hb_schema() -> pa.schema:
    return pa.schema(
        [
            _PK_HASH_COLUMN_FIELD,
            _ORDERED_RECORD_IDX_COLUMN_FIELD,
            _ORDERED_FILE_IDX_COLUMN_FIELD,
            _PARTITION_STREAM_POSITION_COLUMN_FIELD,
            _DELTA_TYPE_COLUMN_FIELD,
            _IS_SOURCE_COLUMN_FIELD,
        ]
    )



from enum import Enum
from typing import Callable, Dict, Type, Union

import numpy as np
import pandas as pd
import pyarrow as pa
from ray.data.dataset import Dataset
from ray.data.read_api import (
    from_arrow,
    from_arrow_refs,
    from_numpy,
    from_pandas,
    from_pandas_refs,
)

import deltacat.storage as dcs
from deltacat.types.media import TableType
from deltacat.utils import numpy as np_utils
from deltacat.utils import pandas as pd_utils
from deltacat.utils import pyarrow as pa_utils
from deltacat.utils.ray_utils import dataset as ds_utils

TABLE_TYPE_TO_READER_FUNC: Dict[int, Callable] = {
    TableType.PYARROW.value: pa_utils.s3_file_to_table,
    TableType.PANDAS.value: pd_utils.s3_file_to_dataframe,
    TableType.NUMPY.value: np_utils.s3_file_to_ndarray,
}

TABLE_CLASS_TO_WRITER_FUNC: Dict[
    Type[Union[dcs.LocalTable, dcs.DistributedDataset]], Callable
] = {
    pa.Table: pa_utils.table_to_file,
    pd.DataFrame: pd_utils.dataframe_to_file,
    np.ndarray: np_utils.ndarray_to_file,
    Dataset: ds_utils.dataset_to_file,
}

TABLE_CLASS_TO_SLICER_FUNC: Dict[
    Type[Union[dcs.LocalTable, dcs.DistributedDataset]], Callable
] = {
    pa.Table: pa_utils.slice_table,
    pd.DataFrame: pd_utils.slice_dataframe,
    np.ndarray: np_utils.slice_ndarray,
    Dataset: ds_utils.slice_dataset,
}

TABLE_CLASS_TO_SIZE_FUNC: Dict[
    Type[Union[dcs.LocalTable, dcs.DistributedDataset]], Callable
] = {
    pa.Table: pa_utils.table_size,
    pd.DataFrame: pd_utils.dataframe_size,
    np.ndarray: np_utils.ndarray_size,
    Dataset: ds_utils.dataset_size,
}

TABLE_CLASS_TO_TABLE_TYPE: Dict[Type[dcs.LocalTable], str] = {
    pa.Table: TableType.PYARROW.value,
    pd.DataFrame: TableType.PANDAS.value,
    np.ndarray: TableType.NUMPY.value,
}

TABLE_TYPE_TO_DATASET_CREATE_FUNC: Dict[str, Callable] = {
    TableType.PYARROW.value: from_arrow,
    TableType.NUMPY.value: from_numpy,
    TableType.PANDAS.value: from_pandas,
}

TABLE_TYPE_TO_DATASET_CREATE_FUNC_REFS: Dict[str, Callable] = {
    TableType.PYARROW.value: from_arrow_refs,
    TableType.NUMPY.value: from_numpy,
    TableType.PANDAS.value: from_pandas_refs,
}


class TableWriteMode(str, Enum):
    """
    Enum controlling how a given dataset will be written to a table.

    AUTO: CREATE if the table doesn't exist, APPEND if the table exists
    without primary keys, and MERGE if the table exists with primary keys.
    CREATE: Create the table if it doesn't exist, throw an error if it does.
    APPEND: Append to the table if it exists, throw an error if it doesn't.
    REPLACE: Replace existing table contents with the data to write.
    MERGE: Insert, update, or delete records matching a given predicate.
    Updates or inserts records based on the table's primary and sort keys by
    default.
    """

    AUTO = "auto"
    CREATE = "create"
    APPEND = "append"
    REPLACE = "replace"
    MERGE = "merge"


def get_table_length(table: Union[dcs.LocalTable, dcs.DistributedDataset]) -> int:
    return len(table) if not isinstance(table, Dataset) else table.count()


def get_table_writer(table: Union[dcs.LocalTable, dcs.DistributedDataset]) -> Callable:
    table_writer_func = TABLE_CLASS_TO_WRITER_FUNC.get(type(table))
    if table_writer_func is None:
        msg = (
            f"No writer found for table type: {type(table)}.\n"
            f"Known table types: {TABLE_CLASS_TO_WRITER_FUNC.keys}"
        )
        raise ValueError(msg)
    return table_writer_func


def get_table_slicer(table: Union[dcs.LocalTable, dcs.DistributedDataset]) -> Callable:
    table_slicer_func = TABLE_CLASS_TO_SLICER_FUNC.get(type(table))
    if table_slicer_func is None:
        msg = (
            f"No slicer found for table type: {type(table)}.\n"
            f"Known table types: {TABLE_CLASS_TO_SLICER_FUNC.keys}"
        )
        raise ValueError(msg)
    return table_slicer_func



import hashlib
import os
import time
from typing import Any, Dict


def env_bool(key: str, default: bool) -> int:
    if key in os.environ:
        return bool(os.environ[key])
    return default


def env_integer(key: str, default: int) -> int:
    if key in os.environ:
        return int(os.environ[key])
    return default


def env_string(key: str, default: str) -> str:
    if key in os.environ:
        return os.environ[key]
    return default


def current_time_ms() -> int:
    return int(round(time.time() * 1000))


def sha1_digest(_bytes) -> bytes:
    hasher = hashlib.sha1()
    hasher.update(_bytes)
    return hasher.digest()


def sha1_hexdigest(_bytes) -> str:
    hasher = hashlib.sha1()
    hasher.update(_bytes)
    return hasher.hexdigest()


class ContentTypeKwargsProvider:
    """Abstract callable that takes a content type and keyword arg dictionary
    as input, and returns finalized keyword args as output. Useful for merging
    content-type-specific keyword arguments into an existing fixed dictionary
    of keyword arguments."""

    def _get_kwargs(self, content_type: str, kwargs: Dict[str, Any]) -> Dict[str, Any]:
        raise NotImplementedError

    def __call__(self, content_type: str, kwargs: Dict[str, Any]) -> Dict[str, Any]:
        return self._get_kwargs(content_type, kwargs)


ReadKwargsProvider = ContentTypeKwargsProvider



import logging
from collections import defaultdict
from typing import Any, Dict, List, Optional, Tuple

import numpy as np
import pyarrow as pa
import pyarrow.compute as pc
import ray
from ray import cloudpickle
from ray.types import ObjectRef

from deltacat import logs
from deltacat.compute.compactor import (
    DeltaFileEnvelope,
    DeltaFileLocator,
    PrimaryKeyIndexVersionLocator,
    PyArrowWriteResult,
    RoundCompletionInfo,
    SortKey,
    SortOrder,
)
from deltacat.compute.compactor.utils import primary_key_index as pki
from deltacat.compute.compactor.utils import system_columns as sc
from deltacat.compute.compactor.utils.system_columns import get_minimal_hb_schema
from deltacat.utils.performance import timed_invocation
from deltacat.utils.pyarrow import ReadKwargsProviderPyArrowSchemaOverride

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


MaterializeBucketIndex = int
DeltaFileLocatorToRecords = Dict[DeltaFileLocator, np.ndarray]
DedupeTaskIndex, PickledObjectRef = int, str
DedupeTaskIndexWithObjectId = Tuple[DedupeTaskIndex, PickledObjectRef]
DedupeResult = Tuple[
    Dict[MaterializeBucketIndex, DedupeTaskIndexWithObjectId],
    List[ObjectRef[DeltaFileLocatorToRecords]],
    PyArrowWriteResult,
]


def _union_primary_key_indices(
    s3_bucket: str,
    round_completion_info: RoundCompletionInfo,
    hash_bucket_index: int,
    df_envelopes_list: List[List[DeltaFileEnvelope]],
) -> pa.Table:

    logger.info(
        f"[Hash bucket index {hash_bucket_index}] Reading dedupe input for "
        f"{len(df_envelopes_list)} delta file envelope lists..."
    )
    # read compacted input parquet files first
    # (which implicitly have older stream positions than deltas)
    hb_tables = []
    if round_completion_info:
        tables = pki.download_hash_bucket_entries(
            s3_bucket,
            hash_bucket_index,
            round_completion_info.primary_key_index_version_locator,
            # Enforce consistent column ordering by reading from a schema, to prevent schema mismatch errors
            file_reader_kwargs_provider=ReadKwargsProviderPyArrowSchemaOverride(
                schema=get_minimal_hb_schema()
            ),
        )
        if tables:
            prior_pk_index_table = pa.concat_tables(tables)
            logger.info(
                f"Number of records in prior primary index for hash bucket"
                f" {hash_bucket_index}: {prior_pk_index_table.num_rows}"
            )
            hb_tables.append(prior_pk_index_table)

    # sort by delta file stream position now instead of sorting every row later
    df_envelopes = [d for dfe_list in df_envelopes_list for d in dfe_list]
    df_envelopes = sorted(
        df_envelopes,
        key=lambda df: (df.stream_position, df.file_index),
        reverse=False,  # ascending
    )
    for df_envelope in df_envelopes:
        hb_tables.append(sc.project_delta_file_metadata_on_table(df_envelope))

    hb_table = pa.concat_tables(hb_tables)

    logger.info(
        f"Total records in hash bucket {hash_bucket_index} is {hb_table.num_rows}"
    )
    return hb_table


def _drop_duplicates_by_primary_key_hash(table: pa.Table) -> pa.Table:
    value_to_last_row_idx = {}

    pk_hash_np = sc.pk_hash_column_np(table)
    op_type_np = sc.delta_type_column_np(table)

    assert len(pk_hash_np) == len(op_type_np), (
        f"Primary key digest column length ({len(pk_hash_np)}) doesn't "
        f"match delta type column length ({len(op_type_np)})."
    )

    # TODO(raghumdani): move the dedupe to C++ using arrow methods or similar.
    row_idx = 0
    pk_op_val_iter = zip(pk_hash_np, op_type_np)
    for (pk_val, op_val) in pk_op_val_iter:

        # operation type is True for `UPSERT` and False for `DELETE`
        if op_val:
            # UPSERT this row
            value_to_last_row_idx[pk_val] = row_idx
        else:
            # DELETE this row
            value_to_last_row_idx.pop(pk_val, None)

        row_idx += 1

    return table.take(list(value_to_last_row_idx.values()))


def _write_new_primary_key_index(
    s3_bucket: str,
    new_primary_key_index_version_locator: PrimaryKeyIndexVersionLocator,
    max_rows_per_index_file: int,
    dedupe_task_index: int,
    deduped_tables: List[Tuple[int, pa.Table]],
) -> PyArrowWriteResult:

    logger.info(
        f"[Dedupe task index {dedupe_task_index}] Writing new deduped primary key index: "
        f"{new_primary_key_index_version_locator}"
    )

    pki_results = []
    for hb_index, table in deduped_tables:
        hb_pki_result = pki.write_primary_key_index_files(
            table,
            new_primary_key_index_version_locator,
            s3_bucket,
            hb_index,
            max_rows_per_index_file,
        )
        pki_results.append(hb_pki_result)

    result = PyArrowWriteResult.union(pki_results)
    logger.info(
        f"[Dedupe task index {dedupe_task_index}] Wrote new deduped primary key index: "
        f"{new_primary_key_index_version_locator}. Result: {result}"
    )
    return result


def delta_file_locator_to_mat_bucket_index(
    df_locator: DeltaFileLocator, materialize_bucket_count: int
) -> int:
    digest = df_locator.digest()
    return int.from_bytes(digest, "big") % materialize_bucket_count


@ray.remote(num_returns=3)
def dedupe(
    compaction_artifact_s3_bucket: str,
    round_completion_info: Optional[RoundCompletionInfo],
    new_primary_key_index_version_locator: PrimaryKeyIndexVersionLocator,
    object_ids: List[Any],
    sort_keys: List[SortKey],
    max_records_per_index_file: int,
    num_materialize_buckets: int,
    dedupe_task_index: int,
    delete_old_primary_key_index: bool,
) -> DedupeResult:

    logger.info(f"[Dedupe task {dedupe_task_index}] Starting dedupe task...")
    # TODO (pdames): mitigate risk of running out of memory here in cases of
    #  severe skew of primary key updates in deltas
    src_file_records_obj_refs = [
        cloudpickle.loads(obj_id_pkl) for obj_id_pkl in object_ids
    ]
    logger.info(
        f"[Dedupe task {dedupe_task_index}] Getting delta file envelope "
        f"groups for {len(src_file_records_obj_refs)} object refs..."
    )

    delta_file_envelope_groups_list = ray.get(src_file_records_obj_refs)
    hb_index_to_delta_file_envelopes_list = defaultdict(list)
    for delta_file_envelope_groups in delta_file_envelope_groups_list:
        for hb_idx, dfes in enumerate(delta_file_envelope_groups):
            if dfes is not None:
                hb_index_to_delta_file_envelopes_list[hb_idx].append(dfes)
    src_file_id_to_row_indices = defaultdict(list)
    deduped_tables = []
    logger.info(
        f"[Dedupe task {dedupe_task_index}] Running {len(hb_index_to_delta_file_envelopes_list)} "
        f"dedupe rounds..."
    )
    for hb_idx, dfe_list in hb_index_to_delta_file_envelopes_list.items():
        logger.info(f"{dedupe_task_index}: union primary keys for hb_index: {hb_idx}")

        table, union_time = timed_invocation(
            func=_union_primary_key_indices,
            s3_bucket=compaction_artifact_s3_bucket,
            round_completion_info=round_completion_info,
            hash_bucket_index=hb_idx,
            df_envelopes_list=dfe_list,
        )
        logger.info(
            f"[Dedupe {dedupe_task_index}] Dedupe round input "
            f"record count: {len(table)}, took {union_time}s"
        )

        # sort by sort keys
        if len(sort_keys):
            # TODO (pdames): convert to O(N) dedupe w/ sort keys
            sort_keys.extend(
                [
                    SortKey.of(
                        sc._PARTITION_STREAM_POSITION_COLUMN_NAME, SortOrder.ASCENDING
                    ),
                    SortKey.of(sc._ORDERED_FILE_IDX_COLUMN_NAME, SortOrder.ASCENDING),
                ]
            )
            table = table.take(pc.sort_indices(table, sort_keys=sort_keys))

        # drop duplicates by primary key hash column
        logger.info(
            f"[Dedupe task index {dedupe_task_index}] Dropping duplicates for {hb_idx}"
        )

        table, drop_time = timed_invocation(
            func=_drop_duplicates_by_primary_key_hash, table=table
        )

        logger.info(
            f"[Dedupe task index {dedupe_task_index}] Dedupe round output "
            f"record count: {len(table)}, took: {drop_time}s"
        )

        deduped_tables.append((hb_idx, table))

        stream_position_col = sc.stream_position_column_np(table)
        file_idx_col = sc.file_index_column_np(table)
        row_idx_col = sc.record_index_column_np(table)
        is_source_col = sc.is_source_column_np(table)
        for row_idx in range(len(table)):
            src_dfl = DeltaFileLocator.of(
                is_source_col[row_idx],
                stream_position_col[row_idx],
                file_idx_col[row_idx],
            )
            # TODO(pdames): merge contiguous record number ranges
            src_file_id_to_row_indices[src_dfl].append(row_idx_col[row_idx])

    logger.info(f"Finished all dedupe rounds...")
    mat_bucket_to_src_file_record_count = defaultdict(dict)
    mat_bucket_to_src_file_records: Dict[
        MaterializeBucketIndex, DeltaFileLocatorToRecords
    ] = defaultdict(dict)
    for src_dfl, src_row_indices in src_file_id_to_row_indices.items():
        mat_bucket = delta_file_locator_to_mat_bucket_index(
            src_dfl,
            num_materialize_buckets,
        )
        mat_bucket_to_src_file_records[mat_bucket][src_dfl] = np.array(
            src_row_indices,
        )
        mat_bucket_to_src_file_record_count[mat_bucket][src_dfl] = len(src_row_indices)

    mat_bucket_to_dd_idx_obj_id: Dict[
        MaterializeBucketIndex, DedupeTaskIndexWithObjectId
    ] = {}
    src_file_records_obj_refs: List[ObjectRef[DeltaFileLocatorToRecords]] = []
    for mat_bucket, src_file_records in mat_bucket_to_src_file_records.items():
        object_ref = ray.put(src_file_records)
        pickled_object_ref = cloudpickle.dumps(object_ref)
        src_file_records_obj_refs.append(pickled_object_ref)
        mat_bucket_to_dd_idx_obj_id[mat_bucket] = (
            dedupe_task_index,
            pickled_object_ref,
        )
        del object_ref
        del pickled_object_ref
    logger.info(
        f"Count of materialize buckets with object refs: "
        f"{len(mat_bucket_to_dd_idx_obj_id)}"
    )

    write_pki_result: PyArrowWriteResult = _write_new_primary_key_index(
        compaction_artifact_s3_bucket,
        new_primary_key_index_version_locator,
        max_records_per_index_file,
        dedupe_task_index,
        deduped_tables,
    )

    if delete_old_primary_key_index:
        pki.delete_primary_key_index_version(
            compaction_artifact_s3_bucket,
            round_completion_info.primary_key_index_version_locator,
        )
    logger.info(f"[Dedupe task index {dedupe_task_index}] Finished dedupe task...")
    return mat_bucket_to_dd_idx_obj_id, src_file_records_obj_refs, write_pki_result



import errno
import logging
import os
import subprocess
from subprocess import run
from typing import Any

import ray
from tenacity import RetryError, Retrying, stop_after_attempt, wait_fixed

from deltacat import logs
from deltacat.compute.metastats.utils.constants import (
    MAX_WORKER_MULTIPLIER,
    WORKER_NODE_OBJECT_STORE_MEMORY_RESERVE_RATIO,
)

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))

RAY_DOWN_DEFAULT_RETRY_ATTEMPTS = 3


def run_cmd_exit_code(cmd: str) -> int:
    logger.info(f"running command {cmd}")
    exit_code = int(os.system(cmd))
    logger.info(f"Got {exit_code} when running {cmd}")


def run_cmd_with_retry(cmd: str) -> None:
    retrying = Retrying(
        wait=wait_fixed(2), stop=stop_after_attempt(RAY_DOWN_DEFAULT_RETRY_ATTEMPTS)
    )
    try:
        retrying(run_cmd_exit_code(cmd))
    except RetryError:
        logger.info(f"{cmd} failed after {RAY_DOWN_DEFAULT_RETRY_ATTEMPTS} retries.")


def run_cmd(cmd: str) -> None:
    result = run(cmd, shell=True, capture_output=True)
    exit_code = int(result.returncode)
    assert exit_code == 0, (
        f"`{cmd}` failed. Exit code: {exit_code} " f"Error Trace: {result.stderr}"
    )


def ray_up(cluster_cfg: str) -> None:
    logger.info(f"Starting Ray cluster '{cluster_cfg}'")
    run_cmd(f"ray up {cluster_cfg} -y --no-config-cache --no-restart")
    logger.info(f"Started Ray cluster '{cluster_cfg}'")


def ray_down(cluster_cfg: str) -> None:
    logger.info(f"Destroying Ray cluster '{cluster_cfg}'")
    run_cmd_with_retry(f"ray down {cluster_cfg} -y")
    logger.info(f"Destroyed Ray cluster '{cluster_cfg}'")


def clean_up_cluster_cfg_file(cluster_cfg) -> None:
    logger.info(f"Removing stats cluster config at: '{cluster_cfg}'")
    run_cmd(f"rm -f {cluster_cfg}")
    logger.info(f"Removed stats cluster config at: '{cluster_cfg}'")


def get_head_node_ip(cluster_cfg: str) -> str:
    logger.info(f"Getting Ray cluster head node IP for '{cluster_cfg}'")
    proc = subprocess.run(
        f"ray get-head-ip {cluster_cfg}",
        shell=True,
        capture_output=True,
        text=True,
        check=True,
    )
    # the head node IP should be the last line printed to stdout
    head_node_ip = proc.stdout.splitlines()[-1]
    logger.info(f"Ray cluster head node IP for '{cluster_cfg}': {head_node_ip}")
    return head_node_ip


def ray_init(host, port) -> Any:
    ray_init_uri = f"ray://{host}:{port}"
    logger.info(f"Connecting Ray Client to '{ray_init_uri}'")
    client = ray.init(ray_init_uri, allow_multiple=True)
    logger.info(f"Connected Ray Client to '{ray_init_uri}'")
    return client


def replace_cluster_cfg_vars(
    partition_canonical_string: str,
    trace_id: str,
    file_path: str,
    min_workers: int,
    head_type: str,
    worker_type: str,
    head_object_store_memory_pct: int,
    worker_object_store_memory_pct: int,
) -> str:

    head_object_store_memory_pct = head_object_store_memory_pct if not None else 30
    worker_object_store_memory_pct = WORKER_NODE_OBJECT_STORE_MEMORY_RESERVE_RATIO * 100

    max_workers = int(min_workers * MAX_WORKER_MULTIPLIER)
    with open(file_path, "r+") as file:
        contents = file.read().replace("{{use-internal-ips}}", "True")
        contents = contents.replace(
            "{{partition_canonical_string}}", partition_canonical_string
        )
        contents = contents.replace("'{{trace_id}}'", trace_id)
        contents = contents.replace("'{{min-workers}}'", str(min_workers))
        contents = contents.replace("'{{max-workers}}'", str(max_workers))
        contents = contents.replace("'{{head-instance-type}}'", head_type)
        contents = contents.replace("'{{worker-instance-type}}'", worker_type)
        contents = contents.replace(
            "'{{head-object-store-memory-pct}}'", str(head_object_store_memory_pct)
        )
        contents = contents.replace(
            "'{{worker-object-store-memory-pct}}'", str(worker_object_store_memory_pct)
        )
    partition_id = partition_canonical_string.split("|")[-1]
    out_file_name = f"{trace_id}-{partition_id}.{os.path.basename(file_path)}"
    out_file_dir = os.path.join(os.path.dirname(file_path), "tmp")
    out_file_path = os.path.join(out_file_dir, out_file_name)
    try:
        os.makedirs(os.path.dirname(out_file_path))
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise
    with open(out_file_path, "w") as output:
        output.write(contents)
    return out_file_path



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from deltacat.compute.stats.models.delta_stats import DeltaStats


class StatsClusterSizeEstimator(dict):
    @staticmethod
    def of(
        memory_per_cpu: int,
        file_count_per_cpu: int,
        total_memory_needed: int,
        total_file_count: int,
    ) -> DeltaStats:
        estimator = StatsClusterSizeEstimator()
        estimator["memory_per_cpu"] = memory_per_cpu
        estimator["file_count_per_cpu"] = file_count_per_cpu
        estimator["total_memory_needed"] = total_memory_needed
        estimator["total_file_count"] = total_file_count
        return estimator

    @property
    def memory_per_cpu(self) -> int:
        """
        Returns a list of stats associated to each column in this delta.
        """
        return self["memory_per_cpu"]

    @property
    def file_count_per_cpu(self) -> int:
        """
        Returns a list of stats associated to each column in this delta.
        """
        return self["file_count_per_cpu"]

    @property
    def total_memory_needed(self) -> int:
        """
        Returns a list of stats associated to each column in this delta.
        """
        return self["total_memory_needed"]

    @property
    def total_file_count(self) -> int:
        """
        Returns a list of stats associated to each column in this delta.
        """
        return self["total_file_count"]

    @staticmethod
    def estimate_cpus_needed(estimator: StatsClusterSizeEstimator):

        # TODO(zyiqin): Current implementation is only for a rough guess using the PYARROW_INFLATION_MULTIPLIER,
        #  note the inflation rate is for content_length to pyarrow_table_bytes for all columns.
        #  The full implementation logic should be like:
        #  1. liner regression with 99 confidence level: pull metastats data for all deltas for this partition if len(datapoints) > 30.
        #  2. if not enough previous stats collected for same partition: Fall back to datapoints for all paritions for same table.
        #  3. If not enough stats collected for this table: use average content length to each content_type and content_encoding inflation rates
        #  4. If not enough stats for this content_type and content_encoding combination: use the basic PYARROW_INFLATION_MULTIPLIER instead.
        # So, only option 4 is implemented here since this pre-requirement for first 3 options are not met for first round of metastats&stats collection.

        min_cpus_based_on_memory = (
            estimator.total_memory_needed // estimator.memory_per_cpu
        ) + 1
        min_cpus_based_on_file_count = (
            estimator.total_file_count // estimator.file_count_per_cpu
        ) + 1
        return max(min_cpus_based_on_memory, min_cpus_based_on_file_count)



import json
import logging
from collections import defaultdict
from typing import Any, Callable, Dict, List, Optional, Tuple

import numpy as np
import pyarrow as pa
import ray
import s3fs
from ray import cloudpickle
from ray.types import ObjectRef

from deltacat import logs
from deltacat.aws import s3u
from deltacat.compute.compactor import (
    PrimaryKeyIndexLocator,
    PrimaryKeyIndexMeta,
    PrimaryKeyIndexVersionLocator,
    PrimaryKeyIndexVersionMeta,
    PyArrowWriteResult,
    RoundCompletionInfo,
)
from deltacat.compute.compactor.steps.rehash import rehash_bucket as rb
from deltacat.compute.compactor.steps.rehash import rewrite_index as ri
from deltacat.compute.compactor.utils import round_completion_file as rcf
from deltacat.compute.compactor.utils import system_columns as sc
from deltacat.constants import PRIMARY_KEY_INDEX_WRITE_BOTO3_CONFIG
from deltacat.storage import Manifest, PartitionLocator
from deltacat.types.media import ContentEncoding, ContentType
from deltacat.types.tables import get_table_slicer, get_table_writer
from deltacat.utils.common import ReadKwargsProvider
from deltacat.utils.ray_utils.concurrency import invoke_parallel

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def rehash(
    options_provider: Callable[[int, Any], Dict[str, Any]],
    s3_bucket: str,
    source_partition_locator: PartitionLocator,
    old_rci: RoundCompletionInfo,
    new_hash_bucket_count: int,
    hash_bucket_index_group_count: int,
    records_per_primary_key_index_file: int,
    delete_old_primary_key_index: bool,
) -> RoundCompletionInfo:

    logger.info(
        f"Rehashing primary key index. Old round completion info: "
        f"{old_rci}. New hash bucket count: {new_hash_bucket_count}"
    )

    # collect old primary key index information
    old_pki_version_locator = old_rci.primary_key_index_version_locator
    old_pkiv_meta = old_pki_version_locator.primary_key_index_version_meta
    old_pki_meta = old_pkiv_meta.primary_key_index_meta
    old_compacted_partition_locator = old_pki_meta.compacted_partition_locator
    if old_pkiv_meta.hash_bucket_count == new_hash_bucket_count:
        raise ValueError(
            f"Primary key index rehash failed. Old hash bucket "
            f"count ({new_hash_bucket_count}) is "
            f"equal to new hash bucket count. Partition: "
            f"{old_compacted_partition_locator}."
        )

    # generate a new unique primary key index version locator to rehash into
    new_pki_meta = PrimaryKeyIndexMeta.of(
        old_compacted_partition_locator,
        old_pki_meta.primary_keys,
        old_pki_meta.sort_keys,
        old_pki_meta.primary_key_index_algorithm_version,
    )
    new_pki_locator = PrimaryKeyIndexLocator.of(new_pki_meta)
    new_pki_version_meta = PrimaryKeyIndexVersionMeta.of(
        new_pki_meta,
        new_hash_bucket_count,
    )
    rehashed_pki_version_locator = PrimaryKeyIndexVersionLocator.generate(
        new_pki_version_meta
    )

    # launch a rehash task for each bucket of the old primary key index version
    old_hash_bucket_count = old_pkiv_meta.hash_bucket_count
    hb_tasks_pending = invoke_parallel(
        items=range(old_hash_bucket_count),
        ray_task=rb.rehash_bucket,
        max_parallelism=None,
        options_provider=options_provider,
        s3_bucket=s3_bucket,
        old_pki_version_locator=old_pki_version_locator,
        num_buckets=new_hash_bucket_count,
        num_groups=hash_bucket_index_group_count,
    )
    logger.info(f"Getting {len(hb_tasks_pending)} rehash bucket results...")
    hb_results = ray.get([t[0] for t in hb_tasks_pending])
    logger.info(f"Got {len(hb_results)} rehash bucket results.")
    all_hash_group_idx_to_obj_id = defaultdict(list)
    for hash_group_idx_to_obj_id in hb_results:
        for hash_group_index, object_id in enumerate(hash_group_idx_to_obj_id):
            if object_id:
                all_hash_group_idx_to_obj_id[hash_group_index].append(object_id)
    hash_group_count = len(all_hash_group_idx_to_obj_id)
    logger.info(f"Rehash bucket groups created: {hash_group_count}")

    # write primary key index files for each rehashed output bucket
    pki_stats_promises = invoke_parallel(
        items=all_hash_group_idx_to_obj_id.values(),
        ray_task=ri.rewrite_index,
        max_parallelism=None,
        options_provider=options_provider,
        s3_bucket=s3_bucket,
        new_primary_key_index_version_locator=rehashed_pki_version_locator,
        max_records_per_index_file=records_per_primary_key_index_file,
    )
    logger.info(f"Getting {len(pki_stats_promises)} rewrite index results...")
    pki_stats = ray.get([t[0] for t in pki_stats_promises])
    logger.info(f"Got {len(pki_stats)} rewrite index results.")

    round_completion_info = RoundCompletionInfo.of(
        old_rci.high_watermark,
        old_rci.compacted_delta_locator,
        old_rci.compacted_pyarrow_write_result,
        PyArrowWriteResult.union(pki_stats),
        old_rci.sort_keys_bit_width,
        rehashed_pki_version_locator,
        old_rci.rebase_source_partition_locator,
    )
    rcf.write_round_completion_file(
        s3_bucket,
        source_partition_locator,
        new_pki_locator.primary_key_index_root_path,
        round_completion_info,
    )
    if delete_old_primary_key_index:
        delete_primary_key_index_version(
            s3_bucket,
            old_pki_version_locator,
        )
    logger.info(
        f"Rehashed primary key index. New round completion info: "
        f"{round_completion_info}."
    )
    return round_completion_info


def download_hash_bucket_entries(
    s3_bucket: str,
    hash_bucket_index: int,
    primary_key_index_version_locator: PrimaryKeyIndexVersionLocator,
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
) -> List[pa.Table]:

    pk_index_manifest_s3_url = (
        primary_key_index_version_locator.get_pkiv_hb_index_manifest_s3_url(
            s3_bucket,
            hash_bucket_index,
        )
    )
    result = s3u.download(pk_index_manifest_s3_url, False)
    logger.info(
        f"Downloading primary key index hash bucket manifest entries: "
        f"{pk_index_manifest_s3_url}. Primary key index version "
        f"locator: {primary_key_index_version_locator}"
    )
    pk_index_manifest = Manifest(json.loads(result["Body"].read().decode("utf-8")))
    tables = s3u.download_manifest_entries(
        pk_index_manifest, file_reader_kwargs_provider=file_reader_kwargs_provider
    )
    if not tables:
        logger.warning(
            f"Primary key index manifest is empty at: "
            f"{pk_index_manifest_s3_url}. Primary key index version "
            f"locator: {primary_key_index_version_locator}"
        )
    return tables


def delete_primary_key_index_version(
    s3_bucket: str, pki_version_locator: PrimaryKeyIndexVersionLocator
) -> None:

    logger.info(f"Deleting primary key index: {pki_version_locator}")
    s3u.delete_files_by_prefix(
        s3_bucket,
        pki_version_locator.primary_key_index_version_root_path,
    )
    logger.info(f"Primary key index deleted: {pki_version_locator}")


def group_record_indices_by_hash_bucket(
    pki_table: pa.Table, num_buckets: int
) -> np.ndarray:

    hash_bucket_to_indices = np.empty([num_buckets], dtype="object")
    record_index = 0
    for digest in sc.pk_hash_column_np(pki_table):
        hash_bucket = pk_digest_to_hash_bucket_index(digest, num_buckets)
        if hash_bucket_to_indices[hash_bucket] is None:
            hash_bucket_to_indices[hash_bucket] = []
        hash_bucket_to_indices[hash_bucket].append(record_index)
        record_index += 1
    return hash_bucket_to_indices


def group_hash_bucket_indices(
    hash_bucket_object_groups: np.ndarray, num_buckets: int, num_groups: int
) -> Tuple[np.ndarray, List[ObjectRef]]:
    """
    Groups all the ObjectRef that belongs to a particular hash bucket group and hash bucket index.
    """

    object_refs = []
    hash_bucket_group_to_obj_id = np.empty([num_groups], dtype="object")

    if hash_bucket_object_groups is None:
        return hash_bucket_group_to_obj_id, object_refs

    hb_group_to_object = np.empty([num_groups], dtype="object")
    for hb_index, obj in enumerate(hash_bucket_object_groups):
        if obj:
            hb_group = hb_index % num_groups
            if hb_group_to_object[hb_group] is None:
                hb_group_to_object[hb_group] = np.empty([num_buckets], dtype="object")
            hb_group_to_object[hb_group][hb_index] = obj

    for hb_group, obj in enumerate(hb_group_to_object):
        if obj is None:
            continue
        obj_ref = ray.put(obj)
        pickled_obj_ref = cloudpickle.dumps(obj_ref)
        object_refs.append(pickled_obj_ref)
        hash_bucket_group_to_obj_id[hb_group] = pickled_obj_ref
        # NOTE: The cloudpickle.dumps API call creates an out of band object reference to the object_ref variable.
        # After pickling, Ray cannot track the serialized copy of the object or determine when the ObjectRef has been deserialized
        # (e.g., if the ObjectRef is deserialized by a non-Ray process).
        # Thus the object_ref cannot be tracked by Ray's distributed reference counter, even if it goes out of scope.
        # The object now has a permanent reference and the data can't be freed from Ray’s object store.
        # Manually deleting the untrackable object references offsets these permanent references and
        # helps to allow these objects to be garbage collected normally.
        del obj_ref
        del pickled_obj_ref
    return hash_bucket_group_to_obj_id, object_refs


def pk_digest_to_hash_bucket_index(digest, num_buckets: int) -> int:
    """
    Deterministically get the hash bucket a particular digest belongs to
    based on number of total hash buckets.
    """

    return int.from_bytes(digest, "big") % num_buckets


def write_primary_key_index_files(
    table: pa.Table,
    primary_key_index_version_locator: PrimaryKeyIndexVersionLocator,
    s3_bucket: str,
    hb_index: int,
    records_per_index_file: int,
) -> PyArrowWriteResult:
    """
    Writes primary key index files for the given hash bucket index out to the
    specified S3 bucket at the path identified by the given primary key index
    version locator. Output is written as 1 or more Parquet files with the
    given maximum number of records per file.

    TODO(raghumdani): Support writing primary key index to any data catalog
    """
    logger.info(
        f"Writing primary key index files for hash bucket {hb_index}. "
        f"Primary key index version locator: "
        f"{primary_key_index_version_locator}."
    )
    s3_file_system = s3fs.S3FileSystem(
        anon=False,
        s3_additional_kwargs={
            "ContentType": ContentType.PARQUET.value,
            "ContentEncoding": ContentEncoding.IDENTITY.value,
        },
        config_kwargs=PRIMARY_KEY_INDEX_WRITE_BOTO3_CONFIG,
    )
    pkiv_hb_index_s3_url_base = (
        primary_key_index_version_locator.get_pkiv_hb_index_s3_url_base(
            s3_bucket, hb_index
        )
    )
    manifest_entries = s3u.upload_sliced_table(
        table,
        pkiv_hb_index_s3_url_base,
        s3_file_system,
        records_per_index_file,
        get_table_writer(table),
        get_table_slicer(table),
    )
    manifest = Manifest.of(manifest_entries)
    pkiv_hb_index_s3_manifest_s3_url = (
        primary_key_index_version_locator.get_pkiv_hb_index_manifest_s3_url(
            s3_bucket, hb_index
        )
    )
    s3u.upload(pkiv_hb_index_s3_manifest_s3_url, str(json.dumps(manifest)))
    result = PyArrowWriteResult.of(
        len(manifest_entries),
        table.nbytes,
        manifest.meta.content_length,
        len(table),
    )
    logger.info(
        f"Wrote primary key index files for hash bucket {hb_index}. "
        f"Primary key index version locator: "
        f"{primary_key_index_version_locator}. Result: {result}"
    )
    return result



import copy
import itertools
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import ray
from ray._private.ray_constants import MIN_RESOURCE_GRANULARITY
from ray.types import ObjectRef

from deltacat.utils.ray_utils.runtime import current_node_resource_key


def invoke_parallel(
    items: Iterable,
    ray_task: Callable,
    *args,
    max_parallelism: Optional[int] = 1000,
    options_provider: Callable[[int, Any], Dict[str, Any]] = None,
    kwargs_provider: Callable[[int, Any], Dict[str, Any]] = None,
    **kwargs,
) -> List[Union[ObjectRef, Tuple[ObjectRef, ...]]]:
    """
    Creates a limited number of parallel remote invocations of the given ray
    task. By default each task is provided an ordered item from the input
    collection as its first argument followed by additional ordered arguments
    and keyword arguments. If `max_parallelism` is not None, then synchronously
    waits to reach <= `max_parallelism` in-flight remote invocations before
    returning.

    Args:
        items: Iterable of items to iterate over (in-order), and
        provide as the first input argument to the given ray task.
        ray_task: Ray task to invoke.
        *args: Ordered input arguments to the Ray task to invoke. Provided
        immediately after the input item if `kwargs_provider` is None. Provided
        as the first input arguments to the Ray task if `kwargs_provider` is
        specified.
        max_parallelism: Maximum parallel remote invocations. Defaults to
        1000. Set to `None` for unlimited max parallel remote invocations.
        options_provider: Callback that takes the current item index and item
        as input and returns `ray.remote` options as output.
        kwargs_provider: Callback that takes the current item index and item as
        input and returns a dictionary of `ray.remote` keyword arguments as
        output. Keyword arguments returned override all **kwargs with the
        same key.
        **kwargs: Keyword arguments to the Ray task to invoke.
    Returns:
        List of Ray object references returned from the submitted tasks.
    """
    if max_parallelism is not None and max_parallelism <= 0:
        raise ValueError(f"Max parallelism ({max_parallelism}) must be > 0.")
    pending_ids = []
    for i, item in enumerate(items):
        if max_parallelism is not None and len(pending_ids) > max_parallelism:
            # Some tasks return multiple values while others have only return one.
            # For multiple return values we flatten the list of pending futures.
            # TODO (pdames): Support tasks with an unbound number of return values.
            if isinstance(pending_ids[0], list):
                ray.wait(
                    list(itertools.chain(*pending_ids)),
                    num_returns=int(
                        len(pending_ids[0]) * (len(pending_ids) - max_parallelism)
                    ),
                )
            else:
                ray.wait(pending_ids, num_returns=len(pending_ids) - max_parallelism)
        opt = {}
        if options_provider:
            opt = options_provider(i, item)
        if not kwargs_provider:
            pending_id = ray_task.options(**opt).remote(item, *args, **kwargs)
        else:
            kwargs_dict = kwargs_provider(i, item)
            kwargs.update(kwargs_dict)
            pending_id = ray_task.options(**opt).remote(*args, **kwargs)
        pending_ids.append(pending_id)
    return pending_ids


def current_node_options_provider(*args, **kwargs) -> Dict[str, Any]:
    """Returns a resource dictionary that can be included with ray remote
    options to pin the task or actor on the current node via:
    `foo.options(current_node_options_provider()).remote()`"""
    return {"resources": {current_node_resource_key(): MIN_RESOURCE_GRANULARITY}}


def round_robin_options_provider(
    i: int,
    item: Any,
    resource_keys: List[str],
    *args,
    resource_amount_provider: Callable[[int], int] = lambda i: MIN_RESOURCE_GRANULARITY,
    **kwargs,
) -> Dict[str, Any]:
    """Returns a resource dictionary that can be included with ray remote
    options to round robin indexed tasks or actors across a list of resource
    keys. For example, the following code round-robins 100 tasks across all
    live cluster nodes:
    ```
    resource_keys = live_node_resource_keys()
    for i in range(100):
        opt = round_robin_options_provider(i, resource_keys=resource_keys)
        foo.options(**opt).remote()
    ```
    """
    opts = kwargs.get("pg_config")
    if opts:
        new_opts = copy.deepcopy(opts)
        bundle_key_index = i % len(
            new_opts["scheduling_strategy"].placement_group.bundle_specs
        )
        new_opts["scheduling_strategy"].placement_group_bundle_index = bundle_key_index
        return new_opts
    else:
        assert resource_keys, f"No resource keys given to round robin!"
        resource_key_index = i % len(resource_keys)
        key = resource_keys[resource_key_index]
        return {"resources": {key: resource_amount_provider(resource_key_index)}}



from typing import List

from deltacat.utils.common import env_integer

BOTO_MAX_RETRIES = env_integer("BOTO_MAX_RETRIES", 15)
TIMEOUT_ERROR_CODES: List[str] = ["ReadTimeoutError", "ConnectTimeoutError"]



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

import numpy as np

from deltacat.storage import DeltaType, LocalTable

DeltaFileEnvelopeGroups = np.ndarray


class DeltaFileEnvelope(dict):
    @staticmethod
    def of(
        stream_position: int, file_index: int, delta_type: DeltaType, table: LocalTable
    ) -> DeltaFileEnvelope:
        """Static factory builder for a Delta File Envelope
        `
        Args:
            stream_position: Stream position of a delta.
            file_index: Manifest file index number of a delta.
            delta_type: A delta type.
            table: The table object that represents the delta file.

        Returns:
            A delta file envelope.

        """
        if stream_position is None:
            raise ValueError("Missing delta file envelope stream position.")
        if file_index is None:
            raise ValueError("Missing delta file envelope file index.")
        if delta_type is None:
            raise ValueError("Missing Delta file envelope delta type.")
        if table is None:
            raise ValueError("Missing Delta file envelope table.")
        delta_file_envelope = DeltaFileEnvelope()
        delta_file_envelope["streamPosition"] = stream_position
        delta_file_envelope["fileIndex"] = file_index
        delta_file_envelope["deltaType"] = delta_type.value
        delta_file_envelope["table"] = table
        return delta_file_envelope

    @property
    def stream_position(self) -> int:
        return self["streamPosition"]

    @property
    def file_index(self) -> int:
        return self["fileIndex"]

    @property
    def delta_type(self) -> DeltaType:
        return DeltaType(self["deltaType"])

    @property
    def table(self) -> LocalTable:
        return self["table"]



import unittest
from typing import Tuple

from deltacat.compute.stats.utils.intervals import DeltaRange, merge_intervals


class TestMergeIntervals(unittest.TestCase):
    def test_unbounded_start_range(self):
        intervals = sorted(merge_intervals({(3, 9), (None, 15), (13, 30)}))
        interval: Tuple[DeltaRange, DeltaRange] = intervals[0]
        self.assertEqual(interval[0], None)
        self.assertEqual(interval[1], 30)

    def test_unbounded_end_range(self):
        intervals = sorted(merge_intervals({(3, 9), (2, None), (13, 30)}))
        interval: Tuple[DeltaRange, DeltaRange] = intervals[0]
        self.assertEqual(interval[0], 2)
        self.assertEqual(interval[1], None)

    def test_unbounded_start_end_range(self):
        intervals = sorted(merge_intervals({(None, None)}))
        interval: Tuple[DeltaRange, DeltaRange] = intervals[0]
        self.assertEqual(interval[0], None)
        self.assertEqual(interval[1], None)

    def test_no_overlap_range(self):
        intervals = sorted(merge_intervals({(3, 9), (11, 14), (19, 30)}))
        interval1: Tuple[DeltaRange, DeltaRange] = intervals[0]
        interval2: Tuple[DeltaRange, DeltaRange] = intervals[1]
        interval3: Tuple[DeltaRange, DeltaRange] = intervals[2]
        self.assertEqual(interval1, (3, 9))
        self.assertEqual(interval2, (11, 14))
        self.assertEqual(interval3, (19, 30))

    def test_overlap_range(self):
        intervals = sorted(merge_intervals({(3, 9), (9, 14), (14, 30)}))
        interval1: Tuple[DeltaRange, DeltaRange] = intervals[0]
        self.assertEqual(interval1, (3, 30))

    def test_invalid_range(self):
        self.assertRaises(ValueError, merge_intervals, {(3, 9), (9, 3)})

    def test_invalid_type(self):
        self.assertRaises(ValueError, merge_intervals, {(3, 9), (1.2, 3)})
        self.assertRaises(ValueError, merge_intervals, {(3, 9), ("1", 3)})


if __name__ == "__main__":
    unittest.main()



import logging
from collections import defaultdict
from typing import Dict, List, Optional

import ray
from ray.types import ObjectRef

from deltacat import logs
from deltacat.aws import s3u as s3_utils
from deltacat.aws.clients import client_cache
from deltacat.compute.compactor import DeltaAnnotated
from deltacat.compute.metastats.utils.io import (
    cache_inflation_rate_data_for_delta_stats_ready,
    cache_partition_stats_to_s3,
    collect_stats_by_columns,
)
from deltacat.compute.stats.models.delta_column_stats import DeltaColumnStats
from deltacat.compute.stats.models.delta_stats import DeltaStats
from deltacat.compute.stats.models.manifest_entry_stats import ManifestEntryStats
from deltacat.compute.stats.models.stats_result import StatsResult
from deltacat.storage import DeltaLocator, PartitionLocator
from deltacat.storage import interface as unimplemented_deltacat_storage

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))

# TODO: get cpu info from ray.nodes() resource key
DEFAULT_CPUS_STATS_CLUSTER_INSTANCE = 32


def start_stats_collection(
    batched_delta_stats_compute_list: List[DeltaAnnotated],
    columns: List[str],
    stat_results_s3_bucket: Optional[str] = None,
    metastats_results_s3_bucket: Optional[str] = None,
    deltacat_storage=unimplemented_deltacat_storage,
) -> Dict[str, List[DeltaStats]]:
    """Collects statistics on deltas, given a set of delta stream position ranges.
    Example:
        >>> collect(locator, set((1, 5), (4, 8), (13, 16)))
        {
            1: DeltaStats(),  # DeltaStats for stream positions 1 - 8
            13: DeltaStats()  # DeltaStats for stream positions 13 - 16
        }
    Args:
        source_partition_locator: Reference to the partition locator tied to the given delta stream positions
        delta_stream_position_range_set: A set of intervals with an int type representing finite,
            closed bounded values, and a None type representing unbounded infinity.
        columns: Columns can be optionally included to collect stats on specific columns.
            By default, all columns will be calculated.
        stat_results_s3_bucket: Used as a cache file storage for computed delta stats
        metastats_results_s3_bucket: Used as cache file storage for inflation rate meta stats
        deltacat_storage: Client implementation of the DeltaCAT storage interface
    Returns:
        A mapping of stream positions to their corresponding delta stats.
    """
    # TODO: Add CompactionEventDispatcher for stats collection started event
    delta_stats_compute_pending: List[ObjectRef[Dict[str, List[StatsResult, int]]]] = []

    for batched_deltas in batched_delta_stats_compute_list:
        splitted_annotated_deltas = DeltaAnnotated.split(
            batched_deltas, DEFAULT_CPUS_STATS_CLUSTER_INSTANCE
        )
        for splitted_annotated_delta in splitted_annotated_deltas:
            delta_stats_compute_pending.append(
                collect_stats_by_columns.remote(
                    splitted_annotated_delta, columns, deltacat_storage
                )
            )

    column_stats_map = _process_stats(delta_stats_compute_pending)

    if not batched_delta_stats_compute_list:
        logger.info("No new delta need stats collection")
    else:
        (
            delta_stream_range_stats,
            partition_canonical_string,
        ) = resolve_annotated_delta_stats_to_original_deltas_stats(
            column_stats_map, columns, batched_delta_stats_compute_list[0]
        )

        _cache_stats_res_to_s3(
            stat_results_s3_bucket, delta_stream_range_stats, partition_canonical_string
        )

        base_path = s3_utils.parse_s3_url(metastats_results_s3_bucket).url
        inflation_rate_stats_s3_url = f"{base_path}/inflation-rates.json"
        cache_inflation_rate_data_for_delta_stats_ready(
            delta_stream_range_stats, inflation_rate_stats_s3_url, deltacat_storage
        )
        # TODO: Add CompactionEventDispatcher for stats collection completed event
        return delta_stream_range_stats


def _get_account_id() -> str:
    client = client_cache("sts", None)
    account_id = client.get_caller_identity()["Account"]
    return account_id


def _process_stats(
    delta_stats_compute_pending: List[ObjectRef[DeltaStats]],
) -> List[DeltaStats]:
    delta_stats_processed_list: List[DeltaStats] = _resolve_pending_stats(
        delta_stats_compute_pending
    )

    return delta_stats_processed_list


def _resolve_pending_stats(
    delta_stats_pending_list: List[ObjectRef[DeltaStats]],
) -> List[DeltaStats]:
    delta_stats_processed_list: List[DeltaStats] = []

    while delta_stats_pending_list:
        ready, delta_stats_pending_list = ray.wait(delta_stats_pending_list)
        processed_stats_batch: List[DeltaStats] = ray.get(ready)
        delta_stats_processed_list.extend(processed_stats_batch)

    return delta_stats_processed_list


def _cache_stats_res_to_s3(
    stat_results_s3_bucket, delta_stream_range_stats, partition_canonical_string
):
    if stat_results_s3_bucket:
        # Cache the stats into the file store
        cache_partition_stats_to_s3(
            stat_results_s3_bucket, delta_stream_range_stats, partition_canonical_string
        )


def resolve_annotated_delta_stats_to_original_deltas_stats(
    column_stats_map, column_names, delta_annotated
) -> Dict[int, DeltaStats]:

    partition_values = delta_annotated["deltaLocator"]["partitionLocator"][
        "partitionValues"
    ]
    partition_id = delta_annotated["deltaLocator"]["partitionLocator"]["partitionId"]
    stream_locator = delta_annotated["deltaLocator"]["partitionLocator"][
        "streamLocator"
    ]
    partition_locator = PartitionLocator.of(
        stream_locator, partition_values, partition_id
    )

    # Dict[stream_position: List[StatsResult]]
    manifest_column_stats_list = defaultdict(lambda: [])
    for i in range(len(column_stats_map)):
        for column_name in column_names:
            for j in range(len(column_stats_map[i][column_name])):
                manifest_column_stats_list[
                    column_stats_map[i][column_name][j][1]
                ].append([column_stats_map[i][column_name][j][0], column_name])

    stats_res: Dict[int, List[DeltaStats]] = {}
    for key, value in manifest_column_stats_list.items():
        delta_locator = DeltaLocator.of(partition_locator, key)

        # Dict[column_name: List[StatsResult]]
        manifest_stats_list = defaultdict(lambda: [])
        for manifest_stat in value:
            manifest_stats_list[manifest_stat[1]].append(manifest_stat[0])
        delta_ds_column_stats: List[DeltaColumnStats] = []
        for column_name, column_manifest_stats_list in manifest_stats_list.items():

            column_manifest_stats = ManifestEntryStats.of(
                column_manifest_stats_list, delta_locator
            )
            dataset_column_stats = DeltaColumnStats.of(
                column_name, column_manifest_stats
            )
            delta_ds_column_stats.append(dataset_column_stats)

        dataset_stats: DeltaStats = DeltaStats.of(delta_ds_column_stats)
        stats_res[key] = dataset_stats

    return stats_res, partition_locator.canonical_string()



import logging
from functools import lru_cache
from typing import Optional

import boto3
from boto3.exceptions import ResourceNotExistsError
from boto3.resources.base import ServiceResource
from botocore.client import BaseClient
from botocore.config import Config

from deltacat import logs
from deltacat.aws.constants import BOTO_MAX_RETRIES

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def _resource(name: str, region: Optional[str], **kwargs) -> ServiceResource:
    boto_config = Config(retries={"max_attempts": BOTO_MAX_RETRIES, "mode": "standard"})
    return boto3.resource(
        name,
        region,
        config=boto_config,
        **kwargs,
    )


def _client(name: str, region: Optional[str], **kwargs) -> BaseClient:
    try:
        # try to re-use a client from the resource cache first
        return resource_cache(name, region, **kwargs).meta.client
    except ResourceNotExistsError:
        # fall back for clients without an associated resource
        boto_config = Config(
            retries={"max_attempts": BOTO_MAX_RETRIES, "mode": "standard"}
        )
        return boto3.client(
            name,
            region,
            config=boto_config,
            **kwargs,
        )


def resource_cache(name: str, region: Optional[str], **kwargs) -> ServiceResource:
    # we don't use the @lru_cache decorator because Ray can't pickle it
    cached_function = lru_cache()(_resource)
    return cached_function(name, region, **kwargs)


def client_cache(name: str, region: Optional[str], **kwargs) -> BaseClient:
    # we don't use the @lru_cache decorator because Ray can't pickle it
    cached_function = lru_cache()(_client)
    return cached_function(name, region, **kwargs)



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from typing import Any, Dict, Optional

from deltacat.compute.compactor.model.primary_key_index import (
    PrimaryKeyIndexVersionLocator,
)
from deltacat.compute.compactor.model.pyarrow_write_result import PyArrowWriteResult
from deltacat.storage import DeltaLocator, PartitionLocator


class RoundCompletionInfo(dict):
    @staticmethod
    def of(
        high_watermark: int,
        compacted_delta_locator: DeltaLocator,
        compacted_pyarrow_write_result: PyArrowWriteResult,
        pk_index_pyarrow_write_result: PyArrowWriteResult,
        sort_keys_bit_width: int,
        primary_key_index_version_locator: PrimaryKeyIndexVersionLocator,
        rebase_source_partition_locator: Optional[PartitionLocator],
    ) -> RoundCompletionInfo:

        rci = RoundCompletionInfo()
        rci["highWatermark"] = high_watermark
        rci["compactedDeltaLocator"] = compacted_delta_locator
        rci["compactedPyarrowWriteResult"] = compacted_pyarrow_write_result
        rci["pkIndexPyarrowWriteResult"] = pk_index_pyarrow_write_result
        rci["sortKeysBitWidth"] = sort_keys_bit_width
        rci["primaryKeyIndexVersionLocator"] = primary_key_index_version_locator
        rci["rebaseSourcePartitionLocator"] = rebase_source_partition_locator
        return rci

    @property
    def high_watermark(self) -> int:
        return self["highWatermark"]

    @property
    def compacted_delta_locator(self) -> DeltaLocator:
        val: Dict[str, Any] = self.get("compactedDeltaLocator")
        if val is not None and not isinstance(val, DeltaLocator):
            self["compactedDeltaLocator"] = val = DeltaLocator(val)
        return val

    @property
    def compacted_pyarrow_write_result(self) -> PyArrowWriteResult:
        val: Dict[str, Any] = self.get("compactedPyarrowWriteResult")
        if val is not None and not isinstance(val, PyArrowWriteResult):
            self["compactedPyarrowWriteResult"] = val = PyArrowWriteResult(val)
        return val

    @property
    def pk_index_pyarrow_write_result(self) -> PyArrowWriteResult:
        val: Dict[str, Any] = self.get("pkIndexPyarrowWriteResult")
        if val is not None and not isinstance(val, PyArrowWriteResult):
            self["pkIndexPyarrowWriteResult"] = val = PyArrowWriteResult(val)
        return val

    @property
    def sort_keys_bit_width(self) -> int:
        return self["sortKeysBitWidth"]

    @property
    def primary_key_index_version_locator(self) -> PrimaryKeyIndexVersionLocator:
        val: Dict[str, Any] = self.get("primaryKeyIndexVersionLocator")
        if val is not None and not isinstance(val, PrimaryKeyIndexVersionLocator):
            self["primaryKeyIndexVersionLocator"] = val = PrimaryKeyIndexVersionLocator(
                val
            )
        return val

    @property
    def rebase_source_partition_locator(self) -> Optional[PartitionLocator]:
        return self.get("rebaseSourcePartitionLocator")



import functools
import logging
from collections import defaultdict
from typing import Dict, List, Optional, Set, Tuple

import pyarrow as pa
import ray

from deltacat import logs
from deltacat.compute.compactor import (
    PrimaryKeyIndexLocator,
    PrimaryKeyIndexMeta,
    PrimaryKeyIndexVersionLocator,
    PrimaryKeyIndexVersionMeta,
    PyArrowWriteResult,
    RoundCompletionInfo,
    SortKey,
)
from deltacat.compute.compactor.steps import dedupe as dd
from deltacat.compute.compactor.steps import hash_bucket as hb
from deltacat.compute.compactor.steps import materialize as mat
from deltacat.compute.compactor.utils import io
from deltacat.compute.compactor.utils import primary_key_index as pki
from deltacat.compute.compactor.utils import round_completion_file as rcf
from deltacat.compute.stats.models.delta_stats import DeltaStats
from deltacat.storage import Delta, DeltaLocator, Partition, PartitionLocator
from deltacat.storage import interface as unimplemented_deltacat_storage
from deltacat.types.media import ContentType
from deltacat.utils.placement import PlacementGroupConfig
from deltacat.utils.ray_utils.concurrency import (
    invoke_parallel,
    round_robin_options_provider,
)
from deltacat.utils.ray_utils.runtime import live_node_resource_keys

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))

_PRIMARY_KEY_INDEX_ALGORITHM_VERSION: str = "1.0"


def check_preconditions(
    source_partition_locator: PartitionLocator,
    compacted_partition_locator: PartitionLocator,
    sort_keys: List[SortKey],
    max_records_per_output_file: int,
    new_hash_bucket_count: Optional[int],
    deltacat_storage=unimplemented_deltacat_storage,
) -> int:

    assert (
        source_partition_locator.partition_values
        == compacted_partition_locator.partition_values
    ), (
        "In-place compaction must use the same partition values for the "
        "source and destination."
    )
    assert (
        max_records_per_output_file >= 1
    ), "Max records per output file must be a positive value"
    if new_hash_bucket_count is not None:
        assert (
            new_hash_bucket_count >= 1
        ), "New hash bucket count must be a positive value"
    return SortKey.validate_sort_keys(
        source_partition_locator,
        sort_keys,
        deltacat_storage,
    )


def compact_partition(
    source_partition_locator: PartitionLocator,
    destination_partition_locator: PartitionLocator,
    primary_keys: Set[str],
    compaction_artifact_s3_bucket: str,
    last_stream_position_to_compact: int,
    *,
    hash_bucket_count: Optional[int] = None,
    sort_keys: List[SortKey] = None,
    records_per_primary_key_index_file: int = 38_000_000,
    records_per_compacted_file: int = 4_000_000,
    input_deltas_stats: Dict[int, DeltaStats] = None,
    min_pk_index_pa_bytes: int = 0,
    min_hash_bucket_chunk_size: int = 0,
    compacted_file_content_type: ContentType = ContentType.PARQUET,
    delete_prev_primary_key_index: bool = False,
    pg_config: Optional[PlacementGroupConfig] = None,
    schema_on_read: Optional[
        pa.schema
    ] = None,  # TODO (ricmiyam): Remove this and retrieve schema from storage API
    rebase_source_partition_locator: Optional[PartitionLocator] = None,
    rebase_source_partition_high_watermark: Optional[int] = None,
    deltacat_storage=unimplemented_deltacat_storage,
) -> Optional[str]:

    logger.info(f"Starting compaction session for: {source_partition_locator}")
    partition = None
    compaction_rounds_executed = 0
    has_next_compaction_round = True
    new_rcf_s3_url = None
    while has_next_compaction_round:
        (
            has_next_compaction_round,
            new_partition,
            new_rci,
            new_rcf_s3_url,
        ) = _execute_compaction_round(
            source_partition_locator,
            destination_partition_locator,
            primary_keys,
            compaction_artifact_s3_bucket,
            last_stream_position_to_compact,
            hash_bucket_count,
            sort_keys,
            records_per_primary_key_index_file,
            records_per_compacted_file,
            input_deltas_stats,
            min_pk_index_pa_bytes,
            min_hash_bucket_chunk_size,
            compacted_file_content_type,
            delete_prev_primary_key_index,
            pg_config,
            schema_on_read,
            rebase_source_partition_locator,
            rebase_source_partition_high_watermark,
            deltacat_storage,
        )
        if new_partition:
            partition = new_partition
            destination_partition_locator = new_partition.locator
            compaction_rounds_executed += 1
        # Take new primary key index sizes into account for subsequent compaction rounds and their dedupe steps
        if new_rci:
            min_pk_index_pa_bytes = new_rci.pk_index_pyarrow_write_result.pyarrow_bytes

    logger.info(
        f"Partition-{source_partition_locator.partition_values}-> Compaction session data processing completed in "
        f"{compaction_rounds_executed} rounds."
    )
    if partition:
        logger.info(f"Committing compacted partition to: {partition.locator}")
        partition = deltacat_storage.commit_partition(partition)
        logger.info(f"Committed compacted partition: {partition}")
    logger.info(f"Completed compaction session for: {source_partition_locator}")
    return new_rcf_s3_url


@ray.remote(num_cpus=0.1, num_returns=3)
def _execute_compaction_round(
    source_partition_locator: PartitionLocator,
    compacted_partition_locator: PartitionLocator,
    primary_keys: Set[str],
    compaction_artifact_s3_bucket: str,
    last_stream_position_to_compact: int,
    new_hash_bucket_count: Optional[int],
    sort_keys: List[SortKey],
    records_per_primary_key_index_file: int,
    records_per_compacted_file: int,
    input_deltas_stats: Dict[int, DeltaStats],
    min_pk_index_pa_bytes: int,
    min_hash_bucket_chunk_size: int,
    compacted_file_content_type: ContentType,
    delete_prev_primary_key_index: bool,
    pg_config: Optional[PlacementGroupConfig],
    schema_on_read: Optional[pa.schema],
    rebase_source_partition_locator: Optional[PartitionLocator],
    rebase_source_partition_high_watermark: Optional[int],
    deltacat_storage=unimplemented_deltacat_storage,
) -> Tuple[bool, Optional[Partition], Optional[RoundCompletionInfo], Optional[str]]:

    if not primary_keys:
        # TODO (pdames): run simple rebatch to reduce all deltas into 1 delta
        #  with normalized manifest entry sizes
        raise NotImplementedError(
            "Compaction only supports tables with 1 or more primary keys"
        )
    if sort_keys is None:
        sort_keys = []
    # TODO (pdames): detect and handle schema evolution (at least ensure that
    #  we don't recompact simple backwards-compatible changes like constraint
    #  widening and null column additions).
    # TODO (pdames): detect and optimize in-place compaction

    # check preconditions before doing any computationally expensive work
    bit_width_of_sort_keys = check_preconditions(
        source_partition_locator,
        compacted_partition_locator,
        sort_keys,
        records_per_compacted_file,
        new_hash_bucket_count,
        deltacat_storage,
    )

    # sort primary keys to produce the same pk digest regardless of input order
    primary_keys = sorted(primary_keys)

    cluster_resources = ray.cluster_resources()
    logger.info(f"Total cluster resources: {cluster_resources}")
    node_resource_keys = None
    if pg_config:  # use resource in each placement group
        cluster_resources = pg_config.resource
        cluster_cpus = cluster_resources["CPU"]
    else:  # use all cluster resource
        logger.info(f"Available cluster resources: {ray.available_resources()}")
        cluster_cpus = int(cluster_resources["CPU"])
        logger.info(f"Total cluster CPUs: {cluster_cpus}")
        node_resource_keys = live_node_resource_keys()
        logger.info(
            f"Found {len(node_resource_keys)} live cluster nodes: "
            f"{node_resource_keys}"
        )

    # create a remote options provider to round-robin tasks across all nodes or allocated bundles
    logger.info(f"Setting round robin scheduling with node id:{node_resource_keys}")
    round_robin_opt_provider = functools.partial(
        round_robin_options_provider,
        resource_keys=node_resource_keys,
        pg_config=pg_config.opts if pg_config else None,
    )

    # assign a distinct index to each node in the cluster
    # head_node_ip = urllib.request.urlopen(
    #     "http://169.254.169.254/latest/meta-data/local-ipv4"
    # ).read().decode("utf-8")
    # print(f"head node ip: {head_node_ip}")

    # set max task parallelism equal to total cluster CPUs...
    # we assume here that we're running on a fixed-size cluster - this
    # assumption could be removed but we'd still need to know the maximum
    # "safe" number of parallel tasks that our autoscaling cluster could handle
    max_parallelism = int(cluster_cpus)
    logger.info(f"Max parallelism: {max_parallelism}")

    # get the root path of a compatible primary key index for this round
    compatible_primary_key_index_meta = PrimaryKeyIndexMeta.of(
        compacted_partition_locator,
        primary_keys,
        sort_keys,
        _PRIMARY_KEY_INDEX_ALGORITHM_VERSION,
    )
    compatible_primary_key_index_locator = PrimaryKeyIndexLocator.of(
        compatible_primary_key_index_meta
    )
    compatible_primary_key_index_root_path = (
        compatible_primary_key_index_locator.primary_key_index_root_path
    )

    # read the results from any previously completed compaction round that used
    # a compatible primary key index
    round_completion_info = None
    if not rebase_source_partition_locator:
        logger.info(
            f"Reading round completion file for compatible "
            f"primary key index root path: {compatible_primary_key_index_root_path}"
        )
        round_completion_info = rcf.read_round_completion_file(
            compaction_artifact_s3_bucket,
            source_partition_locator,
            compatible_primary_key_index_root_path,
        )
        logger.info(f"Round completion file: {round_completion_info}")

    # read the previous compaction round's hash bucket count, if any
    old_hash_bucket_count = None
    if round_completion_info:
        old_pki_version_locator = (
            round_completion_info.primary_key_index_version_locator
        )
        old_hash_bucket_count = (
            old_pki_version_locator.primary_key_index_version_meta.hash_bucket_count
        )
        min_pk_index_pa_bytes = (
            round_completion_info.pk_index_pyarrow_write_result.pyarrow_bytes
        )
    else:
        logger.info(
            f"No prior round info read. Source partition: "
            f"{source_partition_locator}. Primary key index locator: "
            f"{compatible_primary_key_index_locator}. Rebase source "
            f"partition locator: {rebase_source_partition_locator}"
        )

    # use the new hash bucket count if provided, or fall back to old count
    hash_bucket_count = (
        new_hash_bucket_count
        if new_hash_bucket_count is not None
        else old_hash_bucket_count
    )

    # discover input delta files
    high_watermark = (
        round_completion_info.high_watermark if round_completion_info else None
    )

    input_deltas = io.discover_deltas(
        source_partition_locator,
        high_watermark,
        last_stream_position_to_compact,
        deltacat_storage,
    )

    if not input_deltas:
        logger.info("No input deltas found to compact.")
        return False, None, None, None

    # limit the input deltas to fit on this cluster and convert them to
    # annotated deltas of equivalent size for easy parallel distribution

    (
        uniform_deltas,
        hash_bucket_count,
        last_stream_position_compacted,
    ) = io.limit_input_deltas(
        input_deltas,
        cluster_resources,
        hash_bucket_count,
        min_pk_index_pa_bytes,
        min_hash_bucket_chunk_size,
        input_deltas_stats=input_deltas_stats,
        deltacat_storage=deltacat_storage,
    )

    assert hash_bucket_count is not None and hash_bucket_count > 0, (
        f"Expected hash bucket count to be a positive integer, but found "
        f"`{hash_bucket_count}`"
    )

    # rehash the primary key index if necessary
    if round_completion_info:
        logger.info(f"Round completion file contents: {round_completion_info}")
        # the previous primary key index is compatible with the current, but
        # will need to be rehashed if the hash bucket count has changed
        if hash_bucket_count != old_hash_bucket_count:
            # TODO(draghave): manually test the path after prior primary key
            #   index was already built
            round_completion_info = pki.rehash(
                round_robin_opt_provider,
                compaction_artifact_s3_bucket,
                source_partition_locator,
                round_completion_info,
                hash_bucket_count,
                max_parallelism,
                records_per_primary_key_index_file,
                delete_prev_primary_key_index,
            )

    # parallel step 1:
    # group like primary keys together by hashing them into buckets
    hb_tasks_pending = invoke_parallel(
        items=uniform_deltas,
        ray_task=hb.hash_bucket,
        max_parallelism=max_parallelism,
        options_provider=round_robin_opt_provider,
        primary_keys=primary_keys,
        sort_keys=sort_keys,
        num_buckets=hash_bucket_count,
        num_groups=max_parallelism,
        deltacat_storage=deltacat_storage,
    )
    logger.info(f"Getting {len(hb_tasks_pending)} hash bucket results...")
    hb_results = ray.get([t[0] for t in hb_tasks_pending])
    logger.info(f"Got {len(hb_results)} hash bucket results.")
    all_hash_group_idx_to_obj_id = defaultdict(list)
    for hash_group_idx_to_obj_id in hb_results:
        for hash_group_index, object_id in enumerate(hash_group_idx_to_obj_id):
            if object_id:
                all_hash_group_idx_to_obj_id[hash_group_index].append(object_id)
    hash_group_count = len(all_hash_group_idx_to_obj_id)
    logger.info(f"Hash bucket groups created: {hash_group_count}")

    # TODO (pdames): when resources are freed during the last round of hash
    #  bucketing, start running dedupe tasks that read existing dedupe
    #  output from S3 then wait for hash bucketing to finish before continuing

    # create a new stream for this round
    compacted_stream_locator = compacted_partition_locator.stream_locator
    stream = deltacat_storage.get_stream(
        compacted_stream_locator.namespace,
        compacted_stream_locator.table_name,
        compacted_stream_locator.table_version,
    )
    partition = deltacat_storage.stage_partition(
        stream,
        compacted_partition_locator.partition_values,
    )
    new_compacted_partition_locator = partition.locator

    # generate a new primary key index locator for this round
    new_primary_key_index_meta = PrimaryKeyIndexMeta.of(
        new_compacted_partition_locator,
        primary_keys,
        sort_keys,
        _PRIMARY_KEY_INDEX_ALGORITHM_VERSION,
    )
    new_primary_key_index_locator = PrimaryKeyIndexLocator.of(
        new_primary_key_index_meta
    )
    new_primary_key_index_root_path = (
        new_primary_key_index_locator.primary_key_index_root_path
    )

    # generate a new primary key index version locator for this round
    new_primary_key_index_version_meta = PrimaryKeyIndexVersionMeta.of(
        new_primary_key_index_meta,
        hash_bucket_count,
    )
    new_pki_version_locator = PrimaryKeyIndexVersionLocator.generate(
        new_primary_key_index_version_meta
    )

    # parallel step 2:
    # discover records with duplicate primary keys in each hash bucket, and
    # identify the index of records to keep or drop based on sort keys
    num_materialize_buckets = max_parallelism
    logger.info(f"Materialize Bucket Count: {num_materialize_buckets}")
    dd_tasks_pending = invoke_parallel(
        items=all_hash_group_idx_to_obj_id.values(),
        ray_task=dd.dedupe,
        max_parallelism=max_parallelism,
        options_provider=round_robin_opt_provider,
        kwargs_provider=lambda index, item: {
            "dedupe_task_index": index,
            "object_ids": item,
        },
        compaction_artifact_s3_bucket=compaction_artifact_s3_bucket,
        round_completion_info=round_completion_info,
        new_primary_key_index_version_locator=new_pki_version_locator,
        sort_keys=sort_keys,
        max_records_per_index_file=records_per_primary_key_index_file,
        num_materialize_buckets=num_materialize_buckets,
        delete_old_primary_key_index=delete_prev_primary_key_index,
    )
    logger.info(f"Getting {len(dd_tasks_pending)} dedupe results...")
    dd_results = ray.get([t[0] for t in dd_tasks_pending])
    logger.info(f"Got {len(dd_results)} dedupe results.")
    all_mat_buckets_to_obj_id = defaultdict(list)
    for mat_bucket_idx_to_obj_id in dd_results:
        for (
            bucket_idx,
            dd_task_index_and_object_id_tuple,
        ) in mat_bucket_idx_to_obj_id.items():
            all_mat_buckets_to_obj_id[bucket_idx].append(
                dd_task_index_and_object_id_tuple
            )
    logger.info(f"Getting {len(dd_tasks_pending)} dedupe result stat(s)...")
    pki_stats = ray.get([t[2] for t in dd_tasks_pending])
    logger.info(f"Got {len(pki_stats)} dedupe result stat(s).")
    logger.info(f"Materialize buckets created: " f"{len(all_mat_buckets_to_obj_id)}")

    # TODO(pdames): when resources are freed during the last round of deduping
    #  start running materialize tasks that read materialization source file
    #  tables from S3 then wait for deduping to finish before continuing

    # TODO(pdames): balance inputs to materialization tasks to ensure that each
    #  task has an approximately equal amount of input to materialize

    # TODO(pdames): garbage collect hash bucket output since it's no longer
    #  needed

    # parallel step 3:
    # materialize records to keep by index
    mat_tasks_pending = invoke_parallel(
        items=all_mat_buckets_to_obj_id.items(),
        ray_task=mat.materialize,
        max_parallelism=max_parallelism,
        options_provider=round_robin_opt_provider,
        kwargs_provider=lambda index, mat_bucket_index_to_obj_id: {
            "mat_bucket_index": mat_bucket_index_to_obj_id[0],
            "dedupe_task_idx_and_obj_id_tuples": mat_bucket_index_to_obj_id[1],
        },
        schema=schema_on_read,
        round_completion_info=round_completion_info,
        source_partition_locator=source_partition_locator,
        partition=partition,
        max_records_per_output_file=records_per_compacted_file,
        compacted_file_content_type=compacted_file_content_type,
        deltacat_storage=deltacat_storage,
    )
    logger.info(f"Getting {len(mat_tasks_pending)} materialize result(s)...")
    mat_results = ray.get(mat_tasks_pending)
    logger.info(f"Got {len(mat_results)} materialize result(s).")

    mat_results = sorted(mat_results, key=lambda m: m.task_index)
    deltas = [m.delta for m in mat_results]
    merged_delta = Delta.merge_deltas(deltas)
    compacted_delta = deltacat_storage.commit_delta(merged_delta)
    logger.info(f"Committed compacted delta: {compacted_delta}")

    new_compacted_delta_locator = DeltaLocator.of(
        new_compacted_partition_locator,
        compacted_delta.stream_position,
    )

    rci_high_watermark = (
        rebase_source_partition_high_watermark
        if rebase_source_partition_high_watermark
        else last_stream_position_compacted
    )
    new_round_completion_info = RoundCompletionInfo.of(
        rci_high_watermark,
        new_compacted_delta_locator,
        PyArrowWriteResult.union([m.pyarrow_write_result for m in mat_results]),
        PyArrowWriteResult.union(pki_stats),
        bit_width_of_sort_keys,
        new_pki_version_locator,
        rebase_source_partition_locator
        or round_completion_info.rebase_source_partition_locator,
    )
    rcf_source_partition_locator = (
        rebase_source_partition_locator
        if rebase_source_partition_locator
        else source_partition_locator
    )
    round_completion_file_s3_url = rcf.write_round_completion_file(
        compaction_artifact_s3_bucket,
        rcf_source_partition_locator,
        new_primary_key_index_root_path,
        new_round_completion_info,
    )
    logger.info(
        f"partition-{source_partition_locator.partition_values},"
        f"compacted at: {last_stream_position_compacted},"
        f"last position: {last_stream_position_to_compact}"
    )
    return (
        (last_stream_position_compacted < last_stream_position_to_compact),
        partition,
        new_round_completion_info,
        round_completion_file_s3_url,
    )



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

import itertools
import logging
from typing import Any, Dict, List, Optional
from uuid import uuid4

from deltacat import logs

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


class Manifest(dict):
    @staticmethod
    def _build_manifest(
        meta: Optional[ManifestMeta],
        entries: Optional[ManifestEntryList],
        author: Optional[ManifestAuthor] = None,
        uuid: str = None,
    ) -> Manifest:
        if not uuid:
            uuid = str(uuid4())
        manifest = Manifest()
        manifest["id"] = uuid
        if meta is not None:
            manifest["meta"] = meta
        if entries is not None:
            manifest["entries"] = entries
        if author is not None:
            manifest["author"] = author
        return manifest

    @staticmethod
    def of(
        entries: ManifestEntryList,
        author: Optional[ManifestAuthor] = None,
        uuid: str = None,
    ) -> Manifest:
        if not uuid:
            uuid = str(uuid4())
        total_record_count = 0
        total_content_length = 0
        total_source_content_length = 0
        content_type = None
        content_encoding = None
        if entries:
            content_type = entries[0].meta.content_type
            content_encoding = entries[0].meta.content_encoding
            for entry in entries:
                meta = entry.meta
                if meta.content_type != content_type:
                    content_type = None
                if meta.content_encoding != content_encoding:
                    content_encoding = None
                entry_content_type = meta.content_type
                if entry_content_type != content_type:
                    msg = (
                        f"Expected all manifest entries to have content "
                        f"type '{content_type}' but found "
                        f"'{entry_content_type}'"
                    )
                    raise ValueError(msg)
                entry_content_encoding = meta["content_encoding"]
                if entry_content_encoding != content_encoding:
                    msg = (
                        f"Expected all manifest entries to have content "
                        f"encoding '{content_encoding}' but found "
                        f"'{entry_content_encoding}'"
                    )
                    raise ValueError(msg)
                total_record_count += meta.record_count or 0
                total_content_length += meta.content_length or 0
                total_source_content_length += meta.source_content_length or 0
        meta = ManifestMeta.of(
            total_record_count,
            total_content_length,
            content_type,
            content_encoding,
            total_source_content_length,
        )
        manifest = Manifest._build_manifest(meta, entries, author, uuid)
        return manifest

    @staticmethod
    def merge_manifests(
        manifests: List[Manifest], author: Optional[ManifestAuthor] = None
    ) -> Manifest:
        all_entries = ManifestEntryList(
            itertools.chain(*[m.entries for m in manifests])
        )
        merged_manifest = Manifest.of(all_entries, author)
        return merged_manifest

    @property
    def meta(self) -> Optional[ManifestMeta]:
        val: Dict[str, Any] = self.get("meta")
        if val is not None and not isinstance(val, ManifestMeta):
            self["meta"] = val = ManifestMeta(val)
        return val

    @property
    def entries(self) -> Optional[ManifestEntryList]:
        val: List[ManifestEntry] = self.get("entries")
        if val is not None and not isinstance(val, ManifestEntryList):
            self["entries"] = val = ManifestEntryList.of(val)
        return val

    @property
    def id(self) -> str:
        return self["id"]

    @property
    def author(self) -> Optional[ManifestAuthor]:
        val: Dict[str, Any] = self.get("author")
        if val is not None and not isinstance(val, ManifestAuthor):
            self["author"] = val = ManifestAuthor(val)
        return val


class ManifestMeta(dict):
    @staticmethod
    def of(
        record_count: Optional[int],
        content_length: Optional[int],
        content_type: Optional[str],
        content_encoding: Optional[str],
        source_content_length: Optional[int] = None,
        credentials: Optional[Dict[str, str]] = None,
        content_type_parameters: Optional[List[Dict[str, str]]] = None,
    ) -> ManifestMeta:
        manifest_meta = ManifestMeta()
        if record_count is not None:
            manifest_meta["record_count"] = record_count
        if content_length is not None:
            manifest_meta["content_length"] = content_length
        if source_content_length is not None:
            manifest_meta["source_content_length"] = source_content_length
        if content_type is not None:
            manifest_meta["content_type"] = content_type
        if content_type_parameters is not None:
            manifest_meta["content_type_parameters"] = content_type_parameters
        if content_encoding is not None:
            manifest_meta["content_encoding"] = content_encoding
        if credentials is not None:
            manifest_meta["credentials"] = credentials
        return manifest_meta

    @property
    def record_count(self) -> Optional[int]:
        return self.get("record_count")

    @property
    def content_length(self) -> Optional[int]:
        return self.get("content_length")

    @property
    def content_type(self) -> Optional[str]:
        return self.get("content_type")

    @property
    def content_encoding(self) -> Optional[str]:
        return self.get("content_encoding")

    @property
    def source_content_length(self) -> Optional[int]:
        return self.get("source_content_length")

    @property
    def content_type_parameters(self) -> Optional[List[Dict[str, str]]]:
        return self.get("content_type_parameters")

    @property
    def credentials(self) -> Optional[Dict[str, str]]:
        return self.get("credentials")


class ManifestAuthor(dict):
    @staticmethod
    def of(name: Optional[str], version: Optional[str]) -> ManifestAuthor:
        manifest_author = ManifestAuthor()
        if name is not None:
            manifest_author["name"] = name
        if version is not None:
            manifest_author["version"] = version
        return manifest_author

    @property
    def name(self) -> Optional[str]:
        return self.get("name")

    @property
    def version(self) -> Optional[str]:
        return self.get("version")


class ManifestEntry(dict):
    @staticmethod
    def of(
        url: Optional[str],
        meta: Optional[ManifestMeta],
        mandatory: bool = True,
        uri: Optional[str] = None,
        uuid: Optional[str] = None,
    ) -> ManifestEntry:
        manifest_entry = ManifestEntry()
        if not (uri or url):
            raise ValueError("No URI or URL specified for manifest entry contents.")
        if (uri and url) and (uri != url):
            raise ValueError(f"Manifest entry URI ({uri}) != URL ({url})")
        if url:
            manifest_entry["url"] = manifest_entry["uri"] = url
        elif uri:
            manifest_entry["url"] = manifest_entry["uri"] = uri
        if meta is not None:
            manifest_entry["meta"] = meta
        if mandatory is not None:
            manifest_entry["mandatory"] = mandatory
        if uuid is not None:
            manifest_entry["id"] = uuid
        return manifest_entry

    @staticmethod
    def from_s3_obj_url(
        url: str,
        record_count: int,
        source_content_length: Optional[int] = None,
        **s3_client_kwargs,
    ) -> ManifestEntry:
        from deltacat.aws import s3u as s3_utils

        s3_obj = s3_utils.get_object_at_url(url, **s3_client_kwargs)
        logger.debug(f"Building manifest entry from {url}: {s3_obj}")
        manifest_entry_meta = ManifestMeta.of(
            record_count,
            s3_obj["ContentLength"],
            s3_obj["ContentType"],
            s3_obj["ContentEncoding"],
            source_content_length,
        )
        manifest_entry = ManifestEntry.of(url, manifest_entry_meta)
        return manifest_entry

    @property
    def uri(self) -> Optional[str]:
        return self.get("uri")

    @property
    def url(self) -> Optional[str]:
        return self.get("url")

    @property
    def meta(self) -> Optional[ManifestMeta]:
        val: Dict[str, Any] = self.get("meta")
        if val is not None and not isinstance(val, ManifestMeta):
            self["meta"] = val = ManifestMeta(val)
        return val

    @property
    def mandatory(self) -> bool:
        return self["mandatory"]

    @property
    def id(self) -> Optional[str]:
        return self.get("id")


class ManifestEntryList(List[ManifestEntry]):
    @staticmethod
    def of(entries: List[ManifestEntry]) -> ManifestEntryList:
        manifest_entries = ManifestEntryList()
        for entry in entries:
            if entry is not None and not isinstance(entry, ManifestEntry):
                entry = ManifestEntry(entry)
            manifest_entries.append(entry)
        return manifest_entries

    def __getitem__(self, item):
        val = super().__getitem__(item)
        if val is not None and not isinstance(val, ManifestEntry):
            self[item] = val = ManifestEntry(val)
        return val



# Allow self-referencing Type hints in Python 3.7.
from __future__ import annotations

from deltacat.storage import Stream, Table, TableVersion


class TableDefinition(dict):
    @staticmethod
    def of(
        table: Table, table_version: TableVersion, stream: Stream
    ) -> TableDefinition:
        return TableDefinition(
            {
                "table": table,
                "tableVersion": table_version,
                "stream": stream,
            }
        )

    @property
    def table(self) -> Table:
        return self["table"]

    @property
    def table_version(self) -> TableVersion:
        return self["tableVersion"]

    @property
    def stream(self) -> Stream:
        return self["stream"]



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from typing import Any, Dict

from deltacat.compute.compactor.model.pyarrow_write_result import PyArrowWriteResult
from deltacat.storage import Delta


class MaterializeResult(dict):
    @staticmethod
    def of(
        delta: Delta, task_index: int, pyarrow_write_result: PyArrowWriteResult
    ) -> MaterializeResult:
        materialize_result = MaterializeResult()
        materialize_result["delta"] = delta
        materialize_result["taskIndex"] = task_index
        materialize_result["paWriteResult"] = pyarrow_write_result
        return materialize_result

    @property
    def delta(self) -> Delta:
        val: Dict[str, Any] = self.get("delta")
        if val is not None and not isinstance(val, Delta):
            self["delta"] = val = Delta(val)
        return val

    @property
    def task_index(self) -> int:
        return self["taskIndex"]

    @property
    def pyarrow_write_result(self) -> PyArrowWriteResult:
        val: Dict[str, Any] = self.get("paWriteResult")
        if val is not None and not isinstance(val, PyArrowWriteResult):
            self["paWriteResult"] = val = PyArrowWriteResult(val)
        return val



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

import pyarrow as pa

from deltacat.storage.model.locator import Locator
from deltacat.storage.model.namespace import NamespaceLocator
from deltacat.storage.model.table import TableLocator
from deltacat.types.media import ContentType


class TableVersion(dict):
    @staticmethod
    def of(
        locator: Optional[TableVersionLocator],
        schema: Optional[Union[pa.Schema, str, bytes]],
        partition_keys: Optional[List[Dict[str, Any]]] = None,
        primary_key_columns: Optional[List[str]] = None,
        description: Optional[str] = None,
        properties: Optional[Dict[str, str]] = None,
        content_types: Optional[List[ContentType]] = None,
    ) -> TableVersion:
        table_version = TableVersion()
        table_version.locator = locator
        table_version.schema = schema
        table_version.partition_keys = partition_keys
        table_version.primary_keys = primary_key_columns
        table_version.description = description
        table_version.properties = properties
        table_version.content_types = content_types
        return table_version

    @property
    def locator(self) -> Optional[TableVersionLocator]:
        val: Dict[str, Any] = self.get("tableVersionLocator")
        if val is not None and not isinstance(val, TableVersionLocator):
            self.locator = val = TableVersionLocator(val)
        return val

    @locator.setter
    def locator(self, table_version_locator: Optional[TableVersionLocator]) -> None:
        self["tableVersionLocator"] = table_version_locator

    @property
    def schema(self) -> Optional[Union[pa.Schema, str, bytes]]:
        return self.get("schema")

    @schema.setter
    def schema(self, schema: Optional[Union[pa.Schema, str, bytes]]) -> None:
        self["schema"] = schema

    @property
    def partition_keys(self) -> Optional[List[Dict[str, Any]]]:
        return self.get("partitionKeys")

    @partition_keys.setter
    def partition_keys(self, partition_keys: Optional[List[Dict[str, Any]]]) -> None:
        self["partitionKeys"] = partition_keys

    @property
    def primary_keys(self) -> Optional[List[str]]:
        return self.get("primaryKeys")

    @primary_keys.setter
    def primary_keys(self, primary_keys: Optional[List[str]]) -> None:
        self["primaryKeys"] = primary_keys

    @property
    def description(self) -> Optional[str]:
        return self.get("description")

    @description.setter
    def description(self, description: Optional[str]) -> None:
        self["description"] = description

    @property
    def properties(self) -> Optional[Dict[str, str]]:
        return self.get("properties")

    @properties.setter
    def properties(self, properties: Optional[Dict[str, str]]) -> None:
        self["properties"] = properties

    @property
    def content_types(self) -> Optional[List[ContentType]]:
        content_types = self.get("contentTypes")
        return (
            None
            if content_types is None
            else [None if _ is None else ContentType(_) for _ in content_types]
        )

    @content_types.setter
    def content_types(self, content_types: Optional[List[ContentType]]) -> None:
        self["contentTypes"] = content_types

    @property
    def namespace_locator(self) -> Optional[NamespaceLocator]:
        table_version_locator = self.locator
        if table_version_locator:
            return table_version_locator.namespace_locator
        return None

    @property
    def table_locator(self) -> Optional[TableLocator]:
        table_version_locator = self.locator
        if table_version_locator:
            return table_version_locator.table_locator
        return None

    @property
    def namespace(self) -> Optional[str]:
        table_version_locator = self.locator
        if table_version_locator:
            return table_version_locator.namespace
        return None

    @property
    def table_name(self) -> Optional[str]:
        table_version_locator = self.locator
        if table_version_locator:
            return table_version_locator.table_name
        return None

    @property
    def table_version(self) -> Optional[str]:
        table_version_locator = self.locator
        if table_version_locator:
            return table_version_locator.table_version
        return None

    def is_supported_content_type(self, content_type: ContentType):
        supported_content_types = self.content_types
        return (not supported_content_types) or (
            content_type in supported_content_types
        )


class TableVersionLocator(Locator, dict):
    @staticmethod
    def of(
        table_locator: Optional[TableLocator], table_version: Optional[str]
    ) -> TableVersionLocator:
        table_version_locator = TableVersionLocator()
        table_version_locator.table_locator = table_locator
        table_version_locator.table_version = table_version
        return table_version_locator

    @staticmethod
    def at(
        namespace: Optional[str],
        table_name: Optional[str],
        table_version: Optional[str],
    ) -> TableVersionLocator:
        table_locator = TableLocator.at(namespace, table_name)
        return TableVersionLocator.of(table_locator, table_version)

    @property
    def table_locator(self) -> Optional[TableLocator]:
        val: Dict[str, Any] = self.get("tableLocator")
        if val is not None and not isinstance(val, TableLocator):
            self.table_locator = val = TableLocator(val)
        return val

    @table_locator.setter
    def table_locator(self, table_locator: Optional[TableLocator]) -> None:
        self["tableLocator"] = table_locator

    @property
    def table_version(self) -> Optional[str]:
        return self.get("tableVersion")

    @table_version.setter
    def table_version(self, table_version: Optional[str]) -> None:
        self["tableVersion"] = table_version

    @property
    def namespace_locator(self) -> Optional[NamespaceLocator]:
        table_locator = self.table_locator
        if table_locator:
            return table_locator.namespace_locator
        return None

    @property
    def namespace(self) -> Optional[str]:
        table_locator = self.table_locator
        if table_locator:
            return table_locator.namespace
        return None

    @property
    def table_name(self) -> Optional[str]:
        table_locator = self.table_locator
        if table_locator:
            return table_locator.table_name
        return None

    def canonical_string(self) -> str:
        """
        Returns a unique string for the given locator that can be used
        for equality checks (i.e. two locators are equal if they have
        the same canonical string).
        """
        tl_hexdigest = self.table_locator.hexdigest()
        table_version = self.table_version
        return f"{tl_hexdigest}|{table_version}"



# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations

from typing import Any, Dict, Optional

from deltacat.storage.model.locator import Locator
from deltacat.storage.model.namespace import NamespaceLocator


class Table(dict):
    @staticmethod
    def of(
        locator: Optional[TableLocator],
        permissions: Optional[Dict[str, Any]] = None,
        description: Optional[str] = None,
        properties: Optional[Dict[str, str]] = None,
    ) -> Table:
        table = Table()
        table.locator = locator
        table.permissions = permissions
        table.description = description
        table.properties = properties
        return table

    @property
    def locator(self) -> Optional[TableLocator]:
        val: Dict[str, Any] = self.get("tableLocator")
        if val is not None and not isinstance(val, TableLocator):
            self.locator = val = TableLocator(val)
        return val

    @locator.setter
    def locator(self, table_locator: Optional[TableLocator]) -> None:
        self["tableLocator"] = table_locator

    @property
    def permissions(self) -> Optional[Dict[str, Any]]:
        return self.get("permissions")

    @permissions.setter
    def permissions(self, permissions: Optional[Dict[str, Any]]) -> None:
        self["permissions"] = permissions

    @property
    def description(self) -> Optional[str]:
        return self.get("description")

    @description.setter
    def description(self, description: Optional[str]) -> None:
        self["description"] = description

    @property
    def properties(self) -> Optional[Dict[str, str]]:
        return self.get("properties")

    @properties.setter
    def properties(self, properties: Optional[Dict[str, str]]) -> None:
        self["properties"] = properties

    @property
    def namespace_locator(self) -> Optional[NamespaceLocator]:
        table_locator = self.locator
        if table_locator:
            return table_locator.namespace_locator
        return None

    @property
    def namespace(self) -> Optional[str]:
        table_locator = self.locator
        if table_locator:
            return table_locator.namespace
        return None

    @property
    def table_name(self) -> Optional[str]:
        table_locator = self.locator
        if table_locator:
            return table_locator.table_name
        return None


class TableLocator(Locator, dict):
    @staticmethod
    def of(
        namespace_locator: Optional[NamespaceLocator], table_name: Optional[str]
    ) -> TableLocator:
        table_locator = TableLocator()
        table_locator.namespace_locator = namespace_locator
        table_locator.table_name = table_name
        return table_locator

    @staticmethod
    def at(namespace: Optional[str], table_name: Optional[str]) -> TableLocator:
        namespace_locator = NamespaceLocator.of(namespace)
        return TableLocator.of(namespace_locator, table_name)

    @property
    def namespace_locator(self) -> NamespaceLocator:
        val: Dict[str, Any] = self.get("namespaceLocator")
        if val is not None and not isinstance(val, NamespaceLocator):
            self.namespace_locator = val = NamespaceLocator(val)
        return val

    @namespace_locator.setter
    def namespace_locator(self, namespace_locator: Optional[NamespaceLocator]) -> None:
        self["namespaceLocator"] = namespace_locator

    @property
    def table_name(self) -> Optional[str]:
        return self.get("tableName")

    @table_name.setter
    def table_name(self, table_name: Optional[str]) -> None:
        self["tableName"] = table_name

    @property
    def namespace(self) -> Optional[str]:
        namespace_locator = self.namespace_locator
        if namespace_locator:
            return namespace_locator.namespace
        return None

    def canonical_string(self) -> str:
        """
        Returns a unique string for the given locator that can be used
        for equality checks (i.e. two locators are equal if they have
        the same canonical string).
        """
        nl_hexdigest = self.namespace_locator.hexdigest()
        table_name = self.table_name
        return f"{nl_hexdigest}|{table_name}"



import unittest

import pyarrow as pa

from deltacat.utils.pyarrow import RecordBatchTables


class TestRecordBatchTables(unittest.TestCase):
    def setUp(self) -> None:
        self.column_names = ["pk", "sk"]

    def test_single_table_with_batches_and_remainder(self):
        min_records_batch = 8
        bt = RecordBatchTables(min_records_batch)
        col1 = pa.array([i for i in range(10)])
        col2 = pa.array(["test"] * 10)
        test_table = pa.Table.from_arrays([col1, col2], names=self.column_names)
        bt.append(test_table)
        self.assertTrue(bt.has_batches())
        self.assertEqual(bt.batched_record_count, 8)
        self.assertTrue(_is_gte_batch_size_and_divisible(bt, min_records_batch))
        self.assertTrue(bt.has_remaining())
        self.assertEqual(bt.remaining_record_count, 2)
        self.assertTrue(_is_sorted(bt, self.column_names[0]))

    def test_single_table_with_no_remainder(self):
        min_records_batch = 5
        bt = RecordBatchTables(min_records_batch)
        col1 = pa.array([i for i in range(min_records_batch)])
        col2 = pa.array(["test"] * min_records_batch)
        test_table = pa.Table.from_arrays([col1, col2], names=self.column_names)
        bt.append(test_table)
        self.assertFalse(bt.has_remaining())
        self.assertTrue(_is_sorted(bt, self.column_names[0]))

    def test_single_table_with_only_batches(self):
        min_records_batch = 10
        bt = RecordBatchTables(min_records_batch)
        col1 = pa.array([i for i in range(min_records_batch)])
        col2 = pa.array(["test"] * min_records_batch)
        test_table = pa.Table.from_arrays([col1, col2], names=self.column_names)
        bt.append(test_table)
        self.assertTrue(bt.has_batches())
        self.assertTrue(_is_gte_batch_size_and_divisible(bt, min_records_batch))
        self.assertFalse(bt.has_remaining())
        self.assertEqual(bt.batched_record_count, 10)
        self.assertEqual(bt.remaining_record_count, 0)
        self.assertTrue(_is_sorted(bt, self.column_names[0]))

    def test_single_table_with_only_remainder(self):
        min_records_batch = 11
        bt = RecordBatchTables(min_records_batch)
        col1 = pa.array([i for i in range(10)])
        col2 = pa.array(["test"] * 10)
        test_table = pa.Table.from_arrays([col1, col2], names=self.column_names)
        bt.append(test_table)
        self.assertFalse(bt.has_batches())
        self.assertTrue(bt.has_remaining())
        self.assertEqual(bt.batched_record_count, 0)
        self.assertEqual(bt.remaining_record_count, 10)
        self.assertTrue(_is_sorted(bt, self.column_names[0]))

    def test_grouped_tables_with_only_remainder(self):
        min_records_batch = 600
        test_table_num_records = 100
        grouped_tables = [
            pa.Table.from_arrays(
                [
                    pa.array(
                        [
                            i
                            for i in range(
                                i * test_table_num_records,
                                (i + 1) * test_table_num_records,
                            )
                        ]
                    ),
                    pa.array(["foo"] * test_table_num_records),
                ],
                names=self.column_names,
            )
            for i in range(5)
        ]

        bt = RecordBatchTables(min_records_batch)
        for table in grouped_tables:
            bt.append(table)
        self.assertFalse(bt.has_batches())
        self.assertTrue(bt.has_remaining())
        self.assertEqual(bt.remaining_record_count, 500)
        self.assertLess(bt.remaining_record_count, min_records_batch)
        self.assertTrue(_is_sorted(bt, self.column_names[0]))

    def test_grouped_tables_with_batches_and_remainder(self):
        min_records_batch = 450
        test_table_num_records = 100
        grouped_tables = [
            pa.Table.from_arrays(
                [
                    pa.array(
                        [
                            i
                            for i in range(
                                i * test_table_num_records,
                                (i + 1) * test_table_num_records,
                            )
                        ]
                    ),
                    pa.array(["foo"] * 100),
                ],
                names=self.column_names,
            )
            for i in range(5)
        ]

        bt = RecordBatchTables(min_records_batch)
        for table in grouped_tables:
            bt.append(table)
        self.assertTrue(bt.has_batches())
        self.assertTrue(_is_gte_batch_size_and_divisible(bt, min_records_batch))
        self.assertTrue(bt.has_remaining())
        self.assertEqual(bt.batched_record_count, 450)
        self.assertEqual(bt.remaining_record_count, 50)
        self.assertTrue(bt.batched_record_count % min_records_batch == 0)
        self.assertLess(bt.remaining_record_count, min_records_batch)
        self.assertTrue(_is_sorted(bt, self.column_names[0]))

    def test_grouped_tables_with_smaller_batch_size_than_table_records(self):
        min_records_batch = 5
        test_table_num_records = 39
        grouped_tables = [
            pa.Table.from_arrays(
                [
                    pa.array(
                        [
                            i
                            for i in range(
                                i * test_table_num_records,
                                (i + 1) * test_table_num_records,
                            )
                        ]
                    ),
                    pa.array(["foo"] * test_table_num_records),
                ],
                names=self.column_names,
            )
            for i in range(3)
        ]

        bt = RecordBatchTables(min_records_batch)
        for table in grouped_tables:
            bt.append(table)
            self.assertTrue(_is_sorted(bt, self.column_names[0]))

        self.assertTrue(bt.has_batches())
        self.assertTrue(_is_gte_batch_size_and_divisible(bt, min_records_batch))
        self.assertEqual(bt.batched_record_count, 115)
        self.assertTrue(bt.batched_record_count % min_records_batch == 0)
        self.assertTrue(bt.has_remaining())
        self.assertEqual(bt.remaining_record_count, 2)
        self.assertLess(bt.remaining_record_count, min_records_batch)
        self.assertTrue(_is_sorted(bt, self.column_names[0]))

    def test_batched_tables_factory_from_input_tables(self):
        min_records_batch = 5
        test_table_num_records = 39
        grouped_tables = [
            pa.Table.from_arrays(
                [
                    pa.array(
                        [
                            i
                            for i in range(
                                i * test_table_num_records,
                                (i + 1) * test_table_num_records,
                            )
                        ]
                    ),
                    pa.array(["foo"] * test_table_num_records),
                ],
                names=self.column_names,
            )
            for i in range(3)
        ]
        bt = RecordBatchTables.from_tables(grouped_tables, min_records_batch)
        self.assertTrue(type(bt), RecordBatchTables)
        self.assertTrue(bt.has_batches())
        self.assertTrue(_is_gte_batch_size_and_divisible(bt, min_records_batch))
        self.assertEqual(bt.batched_record_count, 115)
        self.assertTrue(bt.batched_record_count % min_records_batch == 0)
        self.assertTrue(bt.has_remaining())
        self.assertEqual(bt.remaining_record_count, 2)
        self.assertLess(bt.remaining_record_count, min_records_batch)
        self.assertTrue(_is_sorted(bt, self.column_names[0]))

    def test_clear(self):
        min_records_batch = 8
        bt = RecordBatchTables(min_records_batch)
        col1 = pa.array([i for i in range(10)])
        col2 = pa.array(["test"] * 10)
        test_table = pa.Table.from_arrays([col1, col2], names=self.column_names)
        bt.append(test_table)
        self.assertTrue(bt.has_batches())
        self.assertTrue(_is_gte_batch_size_and_divisible(bt, min_records_batch))
        self.assertEqual(bt.batched_record_count, 8)

        bt.clear_batches()
        self.assertFalse(bt.has_batches())
        self.assertEqual(bt.batched_record_count, 0)

    def test_append_after_clear(self):
        min_records_batch = 8
        bt = RecordBatchTables(min_records_batch)
        col1 = pa.array([i for i in range(10)])
        col2 = pa.array(["test"] * 10)
        test_table = pa.Table.from_arrays([col1, col2], names=self.column_names)
        bt.append(test_table)
        self.assertTrue(bt.has_batches())
        self.assertTrue(_is_gte_batch_size_and_divisible(bt, min_records_batch))
        self.assertEqual(bt.batched_record_count, 8)
        prev_remainder_records = bt.remaining_record_count
        self.assertEqual(bt.remaining_record_count, 2)

        bt.clear_batches()
        self.assertFalse(bt.has_batches())
        self.assertEqual(bt.batched_record_count, 0)

        col1 = pa.array([i for i in range(10, 20)])
        col2 = pa.array(["test"] * 10)
        test_table = pa.Table.from_arrays([col1, col2], names=self.column_names)
        bt.append(test_table)

        self.assertEqual(bt.batched_record_count, 8)
        self.assertEqual(bt.remaining_record_count, 4)
        self.assertNotEquals(prev_remainder_records, bt.remaining_record_count)
        self.assertTrue(_is_sorted(bt, self.column_names[0]))

        bt.clear_remaining()
        self.assertFalse(bt.has_remaining())
        self.assertTrue(bt.remaining_record_count == 0)

    def test_evict(self):
        min_records_batch = 8
        bt = RecordBatchTables(min_records_batch)
        col1 = pa.array([i for i in range(10)])
        col2 = pa.array(["test"] * 10)
        test_table = pa.Table.from_arrays([col1, col2], names=self.column_names)
        bt.append(test_table)
        self.assertTrue(bt.has_batches())
        self.assertTrue(_is_gte_batch_size_and_divisible(bt, min_records_batch))
        self.assertTrue(bt.has_remaining())
        self.assertEqual(bt.batched_record_count, 8)
        self.assertEqual(bt.remaining_record_count, 2)
        prev_batched_records = bt.batched_record_count

        evicted_tables = bt.evict()
        self.assertFalse(bt.has_batches())
        self.assertTrue(bt.has_remaining())
        self.assertEqual(bt.batched_record_count, 0)
        self.assertEqual(bt.remaining_record_count, 2)
        self.assertEqual(sum([len(t) for t in evicted_tables]), prev_batched_records)


def _is_sorted(batched_tables: RecordBatchTables, sort_key: str):
    merged_table = pa.concat_tables(
        [*batched_tables.batched, *batched_tables.remaining]
    )
    explicitly_sorted_merged_table = merged_table.sort_by([(sort_key, "ascending")])
    return explicitly_sorted_merged_table == merged_table


def _is_gte_batch_size_and_divisible(
    batched_tables: RecordBatchTables, min_records_batch: int
):
    return all(
        [
            len(table) // min_records_batch > 0 and len(table) % min_records_batch == 0
            for table in batched_tables.batched
        ]
    )


if __name__ == "__main__":
    unittest.main()



from deltacat.aws.redshift import (
    Manifest,
    ManifestAuthor,
    ManifestEntry,
    ManifestEntryList,
    ManifestMeta,
)
from deltacat.storage.model.delta import Delta, DeltaLocator
from deltacat.storage.model.list_result import ListResult
from deltacat.storage.model.locator import Locator
from deltacat.storage.model.namespace import Namespace, NamespaceLocator
from deltacat.storage.model.partition import Partition, PartitionLocator
from deltacat.storage.model.stream import Stream, StreamLocator
from deltacat.storage.model.table import Table, TableLocator
from deltacat.storage.model.table_version import TableVersion, TableVersionLocator
from deltacat.storage.model.types import (
    CommitState,
    DeltaType,
    DistributedDataset,
    LifecycleState,
    LocalDataset,
    LocalTable,
    SchemaConsistencyType,
)

__all__ = [
    "CommitState",
    "Delta",
    "DeltaLocator",
    "Partition",
    "DeltaType",
    "DistributedDataset",
    "LifecycleState",
    "ListResult",
    "LocalDataset",
    "LocalTable",
    "Locator",
    "Manifest",
    "ManifestAuthor",
    "ManifestEntry",
    "ManifestMeta",
    "ManifestEntryList",
    "Namespace",
    "NamespaceLocator",
    "PartitionLocator",
    "Stream",
    "SchemaConsistencyType",
    "StreamLocator",
    "Table",
    "TableLocator",
    "TableVersion",
    "TableVersionLocator",
]



import logging
import os
import pathlib
from logging import FileHandler, Handler, Logger, LoggerAdapter, handlers
from typing import Union

import ray
from ray.runtime_context import RuntimeContext

from deltacat.constants import APPLICATION_LOG_LEVEL, DELTACAT_LOG_LEVEL

DEFAULT_LOG_LEVEL = "INFO"
DEFAULT_LOG_FORMAT = (
    "%(asctime)s\t%(levelname)s pid=%(process)d %(filename)s:%(lineno)s -- %(message)s"
)
DEFAULT_MAX_BYTES_PER_LOG = 2 ^ 20 * 256  # 256 MiB
DEFAULT_BACKUP_COUNT = 0

DEFAULT_APPLICATION_LOG_DIR = "/tmp/deltacat/var/output/logs/"
DEFAULT_APPLICATION_LOG_BASE_FILE_NAME = "application.info.log"
DEFAULT_DEBUG_APPLICATION_LOG_BASE_FILE_NAME = "application.debug.log"

DEFAULT_DELTACAT_LOG_DIR = "/tmp/deltacat/var/output/logs/"
DEFAULT_DELTACAT_LOG_BASE_FILE_NAME = "deltacat-python.info.log"
DEFAULT_DEBUG_DELTACAT_LOG_BASE_FILE_NAME = "deltacat-python.debug.log"


class RayRuntimeContextLoggerAdapter(logging.LoggerAdapter):
    """
    Logger Adapter for injecting Ray Runtime Context into logging messages.
    """

    def __init__(self, logger: Logger, runtime_context: RuntimeContext):
        super().__init__(logger, {})
        self.runtime_context = runtime_context

    def process(self, msg, kwargs):
        """
        Injects Ray Runtime Context details into each log message.

        This may include information such as the raylet node ID, task/actor ID, job ID,
        placement group ID of the worker, and assigned resources to the task/actor.

        Args:
            msg: The original log message
            kwargs: Keyword arguments for the log message

        Returns: A log message with Ray Runtime Context details

        """
        runtime_context_dict = self.runtime_context.get()
        runtime_context_dict[
            "worker_id"
        ] = self.runtime_context.worker.core_worker.get_worker_id()
        if self.runtime_context.get_task_id() or self.runtime_context.get_actor_id():
            runtime_context_dict[
                "pg_id"
            ] = self.runtime_context.get_placement_group_id()
            runtime_context_dict[
                "assigned_resources"
            ] = self.runtime_context.get_assigned_resources()

        return "(ray_runtime_context=%s) -- %s" % (runtime_context_dict, msg), kwargs

    def __reduce__(self):
        """
        Used to unpickle the class during Ray object store transfer.
        """

        def deserializer(*args):
            return RayRuntimeContextLoggerAdapter(args[0], ray.get_runtime_context())

        return deserializer, (self.logger,)


def _add_logger_handler(logger: Logger, handler: Handler) -> Logger:

    logger.setLevel(logging.getLevelName("DEBUG"))
    logger.addHandler(handler)
    return logger


def _create_rotating_file_handler(
    log_directory: str,
    log_base_file_name: str,
    logging_level: str = DEFAULT_LOG_LEVEL,
    max_bytes_per_log_file: int = DEFAULT_MAX_BYTES_PER_LOG,
    backup_count: int = DEFAULT_BACKUP_COUNT,
    logging_format: str = DEFAULT_LOG_FORMAT,
) -> FileHandler:

    if type(logging_level) is str:
        logging_level = logging.getLevelName(logging_level.upper())
    assert log_base_file_name, "log file name is required"
    assert log_directory, "log directory is required"
    log_dir_path = pathlib.Path(log_directory)
    log_dir_path.mkdir(parents=True, exist_ok=True)
    handler = handlers.RotatingFileHandler(
        os.path.join(log_directory, log_base_file_name),
        maxBytes=max_bytes_per_log_file,
        backupCount=backup_count,
    )
    handler.setFormatter(logging.Formatter(logging_format))
    handler.setLevel(logging_level)
    return handler


def _file_handler_exists(logger: Logger, log_dir: str, log_base_file_name: str) -> bool:

    handler_exists = False
    base_file_path = os.path.join(log_dir, log_base_file_name)
    if len(logger.handlers) > 0:
        norm_base_file_path = os.path.normpath(base_file_path)
        handler_exists = any(
            [
                isinstance(handler, logging.FileHandler)
                and os.path.normpath(handler.baseFilename) == norm_base_file_path
                for handler in logger.handlers
            ]
        )
    return handler_exists


def _configure_logger(
    logger: Logger,
    log_level: str,
    log_dir: str,
    log_base_file_name: str,
    debug_log_base_file_name: str,
) -> Union[Logger, LoggerAdapter]:
    primary_log_level = log_level
    logger.propagate = False
    if log_level.upper() == "DEBUG":
        if not _file_handler_exists(logger, log_dir, debug_log_base_file_name):
            handler = _create_rotating_file_handler(
                log_dir, debug_log_base_file_name, "DEBUG"
            )
            _add_logger_handler(logger, handler)
            primary_log_level = "INFO"
    if not _file_handler_exists(logger, log_dir, log_base_file_name):
        handler = _create_rotating_file_handler(
            log_dir, log_base_file_name, primary_log_level
        )
        _add_logger_handler(logger, handler)
    ray_runtime_ctx = ray.get_runtime_context()
    if ray_runtime_ctx.worker.connected:
        logger = RayRuntimeContextLoggerAdapter(logger, ray_runtime_ctx)

    return logger


def configure_deltacat_logger(logger: Logger) -> Union[Logger, LoggerAdapter]:
    return _configure_logger(
        logger,
        DELTACAT_LOG_LEVEL,
        DEFAULT_DELTACAT_LOG_DIR,
        DEFAULT_DELTACAT_LOG_BASE_FILE_NAME,
        DEFAULT_DEBUG_DELTACAT_LOG_BASE_FILE_NAME,
    )


def configure_application_logger(logger: Logger) -> Union[Logger, LoggerAdapter]:
    return _configure_logger(
        logger,
        APPLICATION_LOG_LEVEL,
        DEFAULT_APPLICATION_LOG_DIR,
        DEFAULT_APPLICATION_LOG_BASE_FILE_NAME,
        DEFAULT_DEBUG_APPLICATION_LOG_BASE_FILE_NAME,
    )
